/*===========================================================================*/

#define EXTENDERVERSION 41000          // DLL version number.  Increment this with each different version

#define INTERFACELEVEL  4              // extender interface level (DON'T CHANGE!)
#define EXTENDER_FEATURE_LEVEL 1004      // extender feature level (DON'T CHANGE)

#define STRICT                         // only if you prototype carefully

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================

                                     Starter.C
                              WIL Extender DLL

                  Copyright 1994-2004 Wilson WindowWare, Inc.

  ---------------------------------------------------------------------------
                                Revision History
  ---------------------------------------------------------------------------



  Rev 31000   Dec 01, 2000   MW
    Hacked Existing Extender to remove all useful code,
	keeping only boilerplate and comments.

    Added a few dummy functions

  Rev 41000   Mar 30, 2004   MW
    Added RETVAR and ANSWERVAR Macros to allow arrays to be returned.
	Changed ramdom comments.

  ---------------------------------------------------------------------------
                                  Description
  ---------------------------------------------------------------------------

  This is an boilerplate of a WIL extender.  A WIL extender is an external DLL
  designed to complement the WIL script processor.  The WIL Processor has
  an "AddExtender" function that will load a WIL extender.  It basically
  performs a Windows API "LoadLibrary" on the extender, and then proceeds
  to extract its capabilities by a series of calls:

    WILExtenderQuery case 1 is designed for interrogation of the extender to
    determine what it is.  This may or may not be used by the calling program.

    Usually the WIL processor first calls the WILQueryExtender function
	case 2 to negotiate a parameter interface level and to decide it the WIL
	engine and the extender DLL are compatible.  Old-style extenders that do
	not support the  WILQueryExtender function are assumed to be interface
	level 1.  This version of the sample supports interface level 2.

    Then, assuming the Extender supports interface level 2, a call to the
    WILExtender2 function, case -1 is performed to determine the number of
    functions, constants, commands and other table items that the extender
    wishes to add to the WIL command table.

    Next a series of calls to the same WILExtender2 function with case -2
    is performed to copy table entries to the main WIL engine tables.

  The WIL interpreter provides a pointer to a 32 bit memory location (LONG
  FAR *).  This pointer provides 32 bits of storage associated with the
  currently running script.  Different scripts will point to a different 32
  bit location.  This allows (if required) state information to be saved by
  the extender.  In addition, the extender may allocate a global memory
  block for storage, and save the handle of the memory block in the memory
  location.  This allows almost unlimited local data for the extender.
  Note that most extenders ignore this field completely.  Just becasue
  it is there does not mean you have to use it.

  When the WIL script processor shuts down, WILExtender2 case -3 will be
  called once to allow the extender to perform clean-up processing.


  The CommandExtTable defines the commands that the extender responds to,
  and the case #'s that correspond with the commands.

  Up to 16 parameters may be passed to the extender.  Each parameter is
  passed as a vipervar structure containing an integer (long), a string, and
  a floating point member, as well as an inittype member indicating what
  type of parameter it actually is (ie, which member(s) of the structure
  contain valid values):

    typedef struct vipervariable
    {
      double  flt;                     // 8-byte floating point number
      LONG    x;                       // if int variable, value is here
      int     inittype;                // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
      union
      {
        LPSTR lpstr;                   // pointer to string in string table
        DWORD handle;                  // handle for miscellaneous object types
      };
    } vipervar, FAR *LPVIPERVAR;

  inittype will be one (or more) of the following:

     VARTYPE_UNDEF         0        // These are bitmapped, var can be both  // undefined (ie, beyond nArgCount)
     VARTYPE_INT           1        // Numeric and string, must check bits   // integer value (in 'x')
     VARTYPE_STRING        2                                                 // string (pointed to by 'lpstr')
     VARTYPE_FILE          5        // FILE and INT bits set                 // FILE and INT bits set (in 'x')
     VARTYPE_OLEOBJECT    17        // OLE AND INT bits set                  // NO LONGER USED
     VARTYPE_FLOATNUM     32        // floating point                        // floating point value (in 'flt')
     VARTYPE_BINARY       65        // BINARY and INT bits set               // BINARY and INT bits set (in 'x')
     VARTYPE_ARRAY       256        // an array                              // an array (complex type using all structure elements)
     VARTYPE_VARIANT     512        // a COM (OLE) variant

             

  Note that a variable can be a string and/or integer and/or float and/or
  VARIANT.  The theory is that if a conversion  had to occur,
  for example, both the alternate values are maintained until the
  varialbe changes.  Use bit checking to see what kind of varialbe you have.

  Your function also receives a pointer to one of these structures to
  optionally use for returning a value to the WIL DLL.  You may use your
  own version of the structure if you wish.

  In order to avoid using C runtime acsii<->string conversion functions, a
  pair of conversion routines (MyLtoa and MyAtol) has been provided.

  You should provide one function to return the version number of your
  extender.  Your version number should start at 41000 and increment as
  the DLL is modified.

  Adopt a naming convention so that all functions in the extender begin
  with the same group of letters.  A function name must be 30 characters or
  less.

  ---------------------------------------------------------------------------

  A number of macros have been provided to assist in returning values to
  the WIL interpreter.  These are:

    RETLONG(num)
      Returns a long integer as the return value of the function.

    RETFLOAT(floatnum)
      Returns a floating point value as the return value of the function.

    RETSTRING(string)
      Returns a string as the return value of the function.

    RETVAR(vv)
      Returns a raw vipervar.  Used for arrays.

    RETERROR(level, rcfilestringid)
      Returns an error level (use -1, -2, or -3).  We recommend -1 and a
      string from the resource file for the error.  The string itself
      should start with a 3-digit error number, followed by a colon.

    RETCANCEL
      Used if you pop up a dialog box and the user hits the Cancel button.

    ANSWERSTRING(string)
      Allows the extender to return a string response, but does not
      return to the WIL interpreter directly as the RETxxx functions do.
      It allows a response to be constucted in a global memory buffer,
      the buffer contents returned, and then allows processing to
      continue, usually to free the global buffer.

    ANSWERLONG
	ANSWERFLOAT
	  Apparently useless functions to allow processing after returning
      the result to the WIL processor.  I imagine that someone somewhere
	  may find a use for them.

   ANSWERVAR
      Allows the extender to return an array response, but does not
      return to the WIL interpreter directly as the RETxxx functions do.
      It allows processing to continue, usually for cleanup processing.


  ---------------------------------------------------------------------------

  One final word of advice for those of you who are new to the joys of
  writing DLL's.  Without getting into a lot of technical detail about
  memory segments and heaps and stacks, let's just say that all programs
  have a data segment (DS, also known as the local heap) and a stack
  segment (SS).  For EXE's, the data segment and stack segment are one and
  the same (DS == SS).  But DLL's don't get their own stack segment; they
  share the stack of the EXE that is calling them, and therefore the data
  segment and stack segment are different (DS != SS).

  That said, you need to keep these important points in mind:

  1. You must ALWAYS use 32-bit pointers to access objects that are on the
     stack, such as arrays and structures.

  2. You have no control over the size of the stack, which you may be
     sharing with other DLL's as well as with your calling EXE.  It's a
     good idea to avoid putting large arrays and the like on the stack.

  3. C Storage Classes 101.  Automatic variables (variables defined within
     functions, also known as local variables) get created on the stack:

       char szBuffer[200];             // goes on the stack

     "Static" variables get created on the heap instead:

       static char szBuffer[200];      // goes on the local heap


  ===========================================================================*/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
                                    Includes
  ---------------------------------------------------------------------------*/

// Standard Windows include
#include <windows.h>

// Special platform-dependent includes
#include "wwwpform.h"
#include "addons.h"

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
                                    Defines
  ---------------------------------------------------------------------------*/


// helpful functions for displaying debugging message boxes
#define MBSS(p1, p2) {MessageBox(GetFocus(), (p2), (p1), MB_OK | MB_SETFOREGROUND);}
#define MBSN(p1, p2) {char szDebug[20]; MyLtoa((p2), szDebug, 10); MessageBox(GetFocus(), szDebug, (p1), MB_OK | MB_SETFOREGROUND);}


// RETLONG passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETLONG(numb)         return (lpViperCallback(((lpViperReturn->x=numb)      +(lpViperReturn->inittype=VARTYPE_INT))      ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETFLOAT passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETFLOAT(numb)        return (lpViperCallback(((lpViperReturn->flt=numb)    +(lpViperReturn->inittype=VARTYPE_FLOATNUM)) ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETSTRING passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETSTRING(str)        return (lpViperCallback(((lpViperReturn->lpstr=str)   +(lpViperReturn->inittype=VARTYPE_STRING))   ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETVAR passes a Vipervar to the WIL interpreter.  It does NO error checking 
// and also returns to the WIL Interpreter.  Generally used to return arrays.

#define RETVAR(p1) {return (lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, (p1)));}



// RETERROR(level,rcfilestringid) passes back an error massage to the WIL
// interpreter.
//   level MUST be -1, -2 or -3.
//     Use -1 for most run of the mill errors.
//         -2 for serious errors
//         -3 for absolutely fatal "we better die" errors
//   rcstringid is the string ID from the RC file that contains the error
//   message.  The first three characters of the string should be the same
//   as the string ID, and the fourth character should be a colon.

//#define RETERROR(sev, errnum) return (lpViperCallback(((LoadMyExtString(hinstDLL, (errnum), lpViperReturn->lpstr = szWork, sizeof(szWork)))+(lpViperReturn->inittype = VARTYPE_STRING)) ?(sev)         :(sev)         , ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETERROR(sev, errnum) return (lpViperCallback(((LoadString(hinstDLL, (errnum), lpViperReturn->lpstr = szWork, sizeof(szWork)))+(lpViperReturn->inittype = VARTYPE_STRING)) ?(sev)         :(sev)         , ViperInterfaceVersion, lpViperStruct, lpViperReturn))


#define RETCANCEL return (-5)

// ANSWERSTRING passes its string argument back as the function return, but
// does not return to the WIL program.  NOTE: The value that is returned by
// ANSWERSTRING must be the value you use in your own 'return' command.

#define ANSWERSTRING(str) lpViperCallback(  ((lpViperReturn->lpstr=str)+(lpViperReturn->inittype=VARTYPE_STRING))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERLONG(numb)  lpViperCallback(  ((lpViperReturn->x=numb)+(lpViperReturn->inittype=VARTYPE_INT))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERFLOAT(numb) lpViperCallback(  ((lpViperReturn->flt=numb)+(lpViperReturn->inittype=VARTYPE_FLOATNUM))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERVAR(p1) {lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, (p1));}



// RETCANCEL returns a Cancel flag
#define RETCANCEL return (-5)


/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
                             Structures / Typedefs
  ---------------------------------------------------------------------------*/

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
                              Function Prototypes
  ---------------------------------------------------------------------------*/

LONG  NEARPASCAL MyAtol(LPSTR);
void  NEARPASCAL MyLtoa(LONG, LPSTR, UINT);


/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
                                Global Variables
  ---------------------------------------------------------------------------*/

/* Superstitious data field to bypass/ignore bugs in certain versions of
   Windows that cause the beginning of the data area to be overwritten.
   This field should be the first data area defined and linked. */
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE hinstDLL;                    // DLL Instance saved
DWORD dwLastError;



/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*---------------------------------------------------------------------------
  Command Extension Table for this DLL

  NOTE:   ALL STRINGS MUST BE RETURNED IN LOWER CASE.  Note also that in our
          table uppercase characters are used, but than in the case -2
          routine, the CharLower function is used to convert them to
          lower case

  NOTE:   The sizes of the string names CAN NOT exceed 30 characters.  Note
          also that in our table the sizes are specified as zero, and that
          the case -2 routine uses lstrlen to compute the size for the return

  Definitions
    ID       Number, starting from 0, incrementing by 1, terminated by -1.
             For functions, it is also the 'case' number used to access the
             function.

    TYPE     Either FUNCTION or CONSTANT.

    MASK     Bitmask indicating the types of the 16 parameters.  Each
             nibble (4 bits) represents one parameter, with the low
             nibble representing param1, and then continuing on.  The
             possible types are:

               0  integer
               1  string
               2  floating point number
               3  array
               F  as-is (don't validate or convert)

    PARAMS/  For constants, this is the actual constant value: long integer
    VALUE    constants only.  For functions, the low byte of the low word
             is the number of required parameters, the high byte of the low
             word is the number of (additional) optional paremeters, and
             the high word is reserved (and should be set to 0).

    SIZE     Set to 0.

    NAME     Name of the function or constant.
             MUST START WITH AN ALPHABETIC CHARACTER.  NO SPACES.
             30 CHARACTERS OR LESS.  USE A NAMING CONVENTION.

  The extender is not called for constants.  Constant value is determined
  and saved during extender initialization.
  ---------------------------------------------------------------------------*/

tablestruct CommandExtTable[] =
{
/*---------------------------------------------------------------------------
  In this example, I've reserved ID's 0 - 9 for constants.  This allows a
  nicer numbering convention for the three-digit error message strings.

  Although it would be nice to define lots of constants for message box
  styles and cursor types and such, don't forget that your user can only
  have a maximum of 500 items, combined, from all the extenders that are
  loaded.  Constants do count towards this limit.  Try to be considerate.

  Also, note that only integer constants are supported.
  ---------------------------------------------------------------------------*/

//----------------------------------------------------------------------
// ID   Type       Mask               Params/Value Size Name
//----------------------------------------------------------------------
  {0,  FUNCTION,  0x0000000000000000, 0x00000000L,  0, "abcLastError"      },  // abcLastError()
  {1,  FUNCTION,  0x0000000000000000, 0x00000001L,  0, "abcExtenderInfo"   },  // abcExtenderInfo(0)
  {2,  FUNCTION,  0x0000000000000011, 0x00000002L,  0, "abcCat2Strings"     },  // abcCat2String(s,s)
  {3,  FUNCTION,  0x0000000000000000, 0x00000002L,  0, "abcAdd2Ints"       },  // abcAdd2Ints(i,i)
  {4,  FUNCTION,  0x0000000000000222, 0x00000003L,  0, "abcAdd3Floats"     },  // abcAdd3Floats(f,f)
//--------------------------------------------------------------------------
  {-1,         0,                   0,  0x00000000,  0, ""             }
};

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================
                                    DllMain
  ===========================================================================*/

BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);
  return (TRUE);
}
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================
                               WILExtenderQuery

  WILExtenderQuery is an required entry point.  It is used
  to provide helpful information to the WIL DLL and possibly other callers.
  ===========================================================================*/

DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*---------------------------------------------------------------------------
  case 1 is used to identify yourself to a caller.

  If lpvInfo is not NULL, it points to a character buffer where you may copy
  a string describing the extender.

  nCount is the size of the buffer pointed to by lpvInfo, in bytes.  When
  called by the WIL DLL, you can expect this to be a 256-byte buffer.

  The return value should be the version number of the extender.
  ---------------------------------------------------------------------------*/
    case 1:                            // version information
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }

      return (EXTENDERVERSION);
    }

/*---------------------------------------------------------------------------
  case 2 is used to inform the WIL DLL of your preferred interface level.

  lpvInfo points to an array of LONG's, indicating the interface levels
  supported by the WIL DLL (each element represents one level).

  nCount indicates the number of elements in the array pointed to by lpvInfo.

  The return value must be the same as one of the elements in the lpvInfo
  array.  Currently, the WIL DLL supports interface levels 1 and 2, so your
  return value should be 1 or 2.  Newly designed extenders (such as WILX)
  use interface level 2 and have a WilExtender2 entry point.  However, it
  is possible to design an extender to support multiple interface levels.

  If you return an invalid value, the extender will not be loaded and the
  AddExtender command will fail.
  ---------------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }

/*---------------------------------------------------------------------------
  Return 0 for any unknown requests, or any requests you don't handle!
  ---------------------------------------------------------------------------*/
    default:
      return (0);
  }
}
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================
                                  WILExtender2

  This is a required entry point.  It is called each time a function in your
  extender is executed (identified by CmdId's >= 0).  In addition, it is
  called to perform various initialization and un-initialization tasks
  (identified by CmdId's < 0).
  ===========================================================================*/

DllExport LONG FAR PASCAL WILExtender2
  (HWND            hWnd,                  // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{
  char szWork[256];                // some local storage,  We need this,


  if (CmdId!=0) dwLastError=0;

  switch (CmdId)                       // the "big switch" (cases < 0 are at the bottom)
  {

/*---------------------------------------------------------------------------
  Case -1 is called once, when your extender is loaded.  It is the place to
  perform your one-time initialization tasks.  The return value should be the
  number of items in your CommandExtTable (functions and constants).  You may
  also return 0 if you are unable to determine the correct number of items.

  You may return a negative value (-1) to indicate that your extender was
  unable to successfully initialize, for example, if it is running on an
  unsupported platform or a dependent DLL could not be loaded.  This will
  cause the AddExtender command to fail.  You may additionally provide a
  string indicating the reason for the failure:

    lpvvArg points to a character buffer into which you may copy an error
    message.

    nArgCount is the size of the buffer pointed to by lpvvArg, in bytes.
    You can expect this to be a 256-byte buffer.
  ---------------------------------------------------------------------------*/
    case -1:                           // Initialization
    {
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1);  // 1 extra at end
    }

/*---------------------------------------------------------------------------
  Case -2 is called repeatedly by the WIL extender.  On each call, one
  CommandTable entry is passed back to the WIL extender.  It is unlikely that
  you will need to modify this code.
  ---------------------------------------------------------------------------*/
    case -2:                           // Loading table
    {
      LPTABLESTRUCT lpTable;
      UINT          j;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      j = (UINT) nArgCount;

      if (CommandExtTable[j].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[j].ident;
        lpTable->type   = CommandExtTable[j].type;
        lpTable->param1 = CommandExtTable[j].param1;
        lpTable->ulMask = CommandExtTable[j].ulMask;
        lpTable->size   = lstrlen((LPSTR) CommandExtTable[j].delim);
        CharLower(CommandExtTable[j].delim);
        lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[j].delim);
      }

      return (1);
    }

/*---------------------------------------------------------------------------
  Case -3 is called during the WIL shutdown process.  If you need to do any
  cleanup processing, do it here.
  ---------------------------------------------------------------------------*/
    case -3:                           // shutdown
    {
      return (1);
    }

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/*---------------------------------------------------------------------------
   Case 0  abcLastError() returns the lasterror lflag which presumably other functions can set
   --------------------------------------------------------------------------*/
	case 0:
		{
			RETLONG(dwLastError);
		}

/*---------------------------------------------------------------------------
   Case  1 andExtenderInfo returns extender info.  As we have nothing else to
   report other than the version, the passed parameter is not checked,
   but someday it might be.
   --------------------------------------------------------------------------*/
	case 1:
		{
			RETLONG(EXTENDERVERSION);
		}

/*---------------------------------------------------------------------------
   Case 2 abcCat2Strings returns the two passed string lstrcat'ed together
   --------------------------------------------------------------------------*/
	case 2:
		{
        GLOBALHANDLE ghString;
		DWORD dwSize;
		LONG retval;

		dwSize=lstrlen( lpvvArg[0].lpstr ) + lstrlen (lpvvArg[1].lpstr) + 2;
		ghString=GlobalAlloc(GPTR,dwSize);
		lstrcpy(ghString, lpvvArg[0].lpstr);
		lstrcat(ghString, lpvvArg[1].lpstr);
        retval = ANSWERSTRING(ghString);
		GlobalFree(ghString);
		return (retval);

		}
/*---------------------------------------------------------------------------
   Case 3 abcAdd2Ints  Adds two integers
   --------------------------------------------------------------------------*/
	case 3:
		{
			RETLONG( lpvvArg[0].x + lpvvArg[1].x);
		}
/*---------------------------------------------------------------------------
   Case 4 abcAdd3Floats
   --------------------------------------------------------------------------*/
	case 4:
		{
			RETFLOAT( lpvvArg[0].flt + lpvvArg[1].flt + lpvvArg[2].flt);
		}






/*---------------------------------------------------------------------------
                                  default case
  ---------------------------------------------------------------------------*/
    default:                           // we shouldn't get here
    {
      RETERROR(-1, 901);               // Unknown function
    }
  }                                    // end of "big switch"
}
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================
                                     MyAtol

  Converts a decimal character string to a long integer.
  ===========================================================================*/

LONG NEARPASCAL MyAtol(LPSTR lpSource)
{
  LPSTR lpSrc;
  LONG  nNum;
  BOOL  bNegate;

  lpSrc = lpSource;
  bNegate = FALSE;
  nNum = 0;

  while (*lpSrc == ' ')
    lpSrc++;

  if (*lpSrc == '-' || *lpSrc == '+')
  {
    if (*lpSrc == '-')
      bNegate = TRUE;

    lpSrc++;
  }

  while (*lpSrc >= '0' && *lpSrc <= '9')
  {
    nNum = nNum * 10 + (*lpSrc - '0');
    lpSrc++;
  }

  if (bNegate)
    nNum = -nNum;

  return (nNum);
}

/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/
/************************************************************************************/

/*===========================================================================
                                     MyLtoa

  Converts a decimal integer to a string in any base (2 >= base >= 36).
  ===========================================================================*/

void NEARPASCAL MyLtoa(LONG nNum, LPSTR lpDest, UINT nBase)
{
  static char szConvert[37] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  static char szNum[30];
  LONG   nTemp;
  UINT   i, j;

  if (nNum == 0x80000000)
  {
    lstrcpy(lpDest, "-2147483648");
    return;
  }

  i = 0;
  j = 29;

  if (nNum < 0)
  {
    lpDest[i++] = '-';
    nNum = -nNum;
  }

  szNum[j--] = 0;

  do
  {
    nTemp = nNum % nBase;
    szNum[j--] = szConvert[(UINT) nTemp];
    nNum = nNum / nBase;
  }
  while(nNum);

  j++;
  lstrcpy(lpDest + i, szNum + j);

  return;
}

