#include "stdafx.h"
#include "stdafx.h"  // Standard MFST Pre-compiled header. Only include stuff that doesn't change much.
using namespace std; // STD/STL. 

#include "resource.h"    // Version and error string stuff.
#include "WTExtender.h"  // The extender template and base class declarations.
#include "MapEx.h"       // The Map example extender declared.

CMapEx MapEx;   // The one and only instance of the extender.

///////////////////////////////////////////////////////////////////////////////////////
// Using the C++ array data type so create the statice array instance and initialize it.
const CWilExtenderT<CMapEx>::EntryT<CMapEx> CWilExtenderT<CMapEx>::m_ExtTable[] =
/////////////////   Member Ptr        Function Name     Name Chars  Required Args Optional Args       Arg Types...
{EntryT<CMapEx>(&CMapEx::mpTest,     _T("mpTest"),         6,            1,           3,           PTInt, PTArr, PTTSt, PTAny),
 EntryT<CMapEx>(&CMapEx::mpSetHash,  _T("mpSetHash"),      9,            1,           0,           PTInt),
 EntryT<CMapEx>(&CMapEx::mpArrayize, _T("mpArrayize"),    10,            1,           0,           PTArr),
 EntryT<CMapEx>(&CMapEx::mpMapize,   _T("mpMapize"),       8,            1,           0,           PTArr) //,
 ////////////////////////////////    Constant Name      Name Chars            Value
 };

//////////////////////////////////////////////////////////////////////////////////////
// Extender WIL function implementations.

/********************************************************************
 * Name: mpTest
 *
 * Purpose: Test the DllMapServices callback function by using each
 *          option with valid input
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "mpTest": Arguments vary with request.
 *
 *                        Return varies with request.
 * 
 * Notes: This function does not perform rigorous testing of the 
 *        DLL's WIL map extender interface although it can be used
 *        to do so.
 *
 ********************************************************************/
LONG CMapEx::mpTest(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	LONG nReturn    = 0;
	LONG nResult    = 0;
	vipervar vMap   = {0}; // Important to initialize because DLL may get cranky if you don't.
	vipervar vKey   = {0};
	vipervar vValue = {0};
	DWORD_PTR dwIterator = 0;
	DWORD_PTR dwCount    = 0;
	int nIterIndex  = 0;
	TCHAR szConversion[64]   = {_T('\0')};

	try {

		// Check the calling DLL's feature level. WIL map support requires feature level 1011.
		if ( !IsInterfaceOk() ) 
			throw CUserException(IDS_ERROR_BADIFACE); // 201040: This extender function requires WIL DLL version 6.21bub or newer
		
		// Call a template method based on DllMapServices request in p1.
		switch( lpvvArg->x ) {
			case 1: 
				// Get the count of key/value pairs in a WIL map. 
				if ( nArgCount < 2 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}
				
				nResult = DllMapCount(lpvvArg+1, &dwCount);
				if ( nResult ) throw CDllMapException(nResult);
				
				{
					CUserLocale Locale;
					if ( _ui64tot_s(dwCount, szConversion, _countof(szConversion), 10) ) {
						nExtError = IDS_SYSTEMERROR;
						break;
					}
				}

				nReturn = ReturnResult(szConversion);  // Return count as WIL string.
				
				break;
			case 3: // Create a map.
				
				nResult = DllMapCreate( &vMap);
				if ( nResult ) throw CDllMapException(nResult);
				
				nReturn =  ReturnResult(&vMap);
				
				break;
			case 4:   // Get value with strings as ANSI.
			case 5:   // Get value with strings as Unicode.
			case 6:   // Get value with strings as-is.
				if ( nArgCount < 3 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}
				if ( !IsString(lpvvArg+2) ) {
					nExtError = IDS_ERROR_STREXPECTED;   // 201033: String expected
					break;
				}

				// Could just pass the request to DllMapServices but this example is also about testing of
				// each template member for correctness.
				if ( lpvvArg->x  == 4 ) // 4 = strings as ANSI
					nResult = DllMapGetANSI(lpvvArg+1, lpvvArg+2, &vValue);
				else if ( lpvvArg->x  == 5 ) // == strings as Unicode.
					nResult = DllMapGetUnicode(lpvvArg+1, lpvvArg+2, &vValue);
				else // 6 = strings as is.
					nResult = DllMapGetAsIs(lpvvArg+1, lpvvArg+2, &vValue );

				if ( nResult < 0 ) {
					nExtError = IDS_ERROR_KEYNOTFOUND; // Key not found
					break;
				}
				if ( nResult > 0) throw CDllMapException(nResult);
				
				nReturn = ReturnResult(&vValue);  // Return a value.

				break;
			case 7: // Add a key value pair.
				
				if ( nArgCount < 4 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}

				if ( !IsString(lpvvArg+2) ) {
					nExtError = IDS_ERROR_STREXPECTED;   // 201033: String expected
					break;
				}

				nResult = DllMapSet(lpvvArg+1, lpvvArg+2, lpvvArg+3);
				if ( nResult ) throw CDllMapException(nResult);
				nReturn = ReturnResult(1);

				break;
			case 8: // Destroy map.
				  
				if ( nArgCount < 2 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}
			
				// IMPORTANT: USE WITH EXTREME CAUTION. FREEING THE MAP CORRUPTS ANY VARIABLES HOLDING REFERENCES 
				//            TO THE MAP.
				nResult = DllMapFree(lpvvArg+1);
				if ( nResult ) throw CDllMapException(nResult);
				nReturn = ReturnResult(1);
				
				break;
			case 9: // Get a map interator.			
				
				if ( nArgCount < 2 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}
				nResult =  DllMapGetIterator(lpvvArg+1, &dwIterator); 
				if ( nResult ) throw CDllMapException(nResult);
				
				// Add to the iterator table.
				nIterIndex = m_Cache.AddIterator(dwIterator);
				if ( !m_Cache.IsIndexInBounds(nIterIndex) ) {
					nExtError = IDS_ERROR_CACHEFULL;     // Too many open map iterator 
					break;
				}

				nReturn = ReturnResult(static_cast <LONG> (nIterIndex));  // return the index to the iterator.
				
				break;
			case 10: // Get ANSI key using an iterator.
			case 11: // Get Unicode key using an iterator.
			case 12: // Get as-is key using an iterator.
				
				if ( nArgCount < 3 ) {
					nExtError = IDS_ERROR_WRONGPARMCNT;  // 201031: Parameter(s) missing
					break;
				}
				if ( !IsMap(lpvvArg+1) ) {
					nExtError = IDS_ERROR_MAPEXPECTED;   // 201032: WIL map expected
					break;
				}

				if ( !StrToInt(GetTstr(lpvvArg+2), 10, &nIterIndex) ) {
					nExtError = IDS_ERROR_INVALIDITER;   // 201035: Invalid WIL map iterator
					break;
				}
				if ( !m_Cache.IsIndexInBounds(nIterIndex) ) {
					nExtError = IDS_ERROR_INVALIDITER;   // 201035: Invalid WIL map iterator
					break;
				}
				
				dwIterator = m_Cache.GetIterator(nIterIndex);
				if ( !dwIterator ) {
					// Assume a zero iterator means that the iteration is done.
					nReturn = ReturnResult(0); // Caller must test for int return type to determine when iteration is complete.
					break;
				}

				if ( lpvvArg->x == 10 )
					nResult =  DllMapGetNextANSI(lpvvArg+1, &vKey, &dwIterator, &vValue); 
				else if ( lpvvArg->x == 11 )
					nResult =  DllMapGetNextUnicode(lpvvArg+1, &vKey, &dwIterator, &vValue); 
				else if ( lpvvArg->x == 12 )
					nResult =  DllMapGetNextAsIs(lpvvArg+1, &vKey, &dwIterator, &vValue); 
				if ( nResult ) throw CDllMapException(nResult);

				// Check for end of iteration.
				if ( dwIterator == 0 ) 
					m_Cache.RemoveIterator(nIterIndex);
				else 
					m_Cache.UpdateIterator(nIterIndex, dwIterator);  // Set next map position.

				nReturn = ReturnResult(&vKey);  // Return the key.
				
				break;
			case 2:  // Setting hash size is supported in separate extender function. "Case 2:" is not necessary but does document what is going on.
			default:
				nExtError = IDS_ERROR_INVALIDOPT;
				break;
		}
	}
	catch(CDllMapException &ex ) { // Dll errors as excptions. (normally not shown to users.)
		nExtError = ex.ResourceId();
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}
	catch(CUserException &eu ) { // User errors as excptions.
		nExtError = eu.ResourceId();
	}

	// Error?
	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	// Success.
	return nReturn;
}

/********************************************************************
 * Name: mpSetHash
 *
 * Purpose: Test the DllMapServices callback function option 2
 *          using a separate function to avoid paramter conversions.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "mpSetHash ": New hash size.
 *
 *                        Return - previous hash size.
 *        
 * Notes:  Exceptions are reserved for unexpected, untoward events and
 *         not used for user input errors.
********************************************************************/
LONG CMapEx::mpSetHash(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	LONG nReturn    = 0;
	LONG nResult    = 0;
	DWORD_PTR dwHashSize = 0;
	

	// Since there are no optional parameters and the first parameter is guaranteed to be an integer
	// no reason to check parameter.
	try {

		// Check the calling DLL's feature level. WIL map support requires feature level 1011.
		if ( !IsInterfaceOk() ) 
			throw CUserException(IDS_ERROR_BADIFACE); // 201040: This extender function requires WIL DLL version 6.21bub or newer

		dwHashSize = static_cast <DWORD_PTR> (lpvvArg->x);
		nResult = DllMapHashSize(&dwHashSize);
		if ( nResult ) throw CDllMapException(nResult);

		nReturn = ReturnResult(static_cast <LONG> (dwHashSize));
	}
	catch(CDllMapException &ex ) { // Dll errors as excptions. (normally not shown to users.)
		nExtError = ex.ResourceId();
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}
	catch(CUserException &eu ) { // User errors as excptions.
		nExtError = eu.ResourceId();
	}

	// Error?
	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	// Success.
	return nReturn;
}

/********************************************************************
 * Name: mpArrayize
 *
 * Purpose: Converts a WIL asociative array to a WIL array.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "mpArrayize": parameter 1 =  WIL map.
 *
 *                        Return - Rank 2 array made up of a map's
 *                                 keys and values. Each key is in 
 *                                 column 0 of each row and the 
 *                                 value for the key is in column 1 
 *                                 of each row.
 *
 * Notes:  Exceptions are reserved for unexpected, untoward events 
 *         not used for user input errors.
 ********************************************************************/
LONG CMapEx::mpArrayize(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError   = 0;
	LONG nReturn     = 0;
	LONG nResult     = 0;
	DWORD_PTR dwSize = 0;
	DWORD_PTR dwPos  = 0;
	vipervar vOut    = {0};
	DWORD dwDims[5]  = {0};

	try {

		// Check the calling DLL's feature level. WIL map support requires feature level 1011.
		if ( !IsInterfaceOk() ) 
			throw CUserException(IDS_ERROR_BADIFACE); // 201040: This extender function requires WIL DLL version 6.21bub or newer
	
		// It's an array but is it a map?
		if ( !IsMap(lpvvArg) ) 
			throw CUserException(IDS_ERROR_MAPEXPECTED); // 201032: WIL map expected

		nResult = DllMapCount(lpvvArg, &dwSize);
		if ( nResult ) throw CDllMapException(nResult);
		if ( dwSize == 0 ) 
			throw CUserException(IDS_ERROR_MAPEMPTY); // 201037: Map with one or more pairs required

		// Create the array used as output.
		dwDims[0] = static_cast <DWORD> (dwSize);
		dwDims[1] = 2;
		nResult = DllArrayServices(&vOut, reinterpret_cast <LPBYTE> (dwDims), sizeof(dwDims), 2);
		if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrayServices"));

		
		// Instantiate and initialize variables used in the following loop.
		vipervar vKey   = {0};
		vipervar vValue = {0};
		DWORD dwSub1    = 0;
		string strKey; 
		string strVal;
		wstring strwVal;
		
		// Prime the iterator pump.
		nResult =  DllMapGetIterator(lpvvArg, &dwPos); 
		if ( nResult ) throw CDllMapException(nResult);
		
		while ( dwPos ) {
			nResult =  DllMapGetNextAsIs(lpvvArg, &vKey, &dwPos, &vValue); 
			if ( nResult ) throw CDllMapException(nResult);
		
			// Add key to the first column of the row.
			// For some reason DllArrHandler does not like to use strings from the WIL string table...
			strKey = vKey.lpstr;
			vKey.lpstr = (char *) strKey.data();  // Oh no! Casting away const. (Its OK because DllArrHandler does not alter the string.)
			nResult = DllArrHandler(&vOut, dwSub1, 0,0,0,0,&vKey, 2);
			if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrHandler"));

			// Add the value to the second column of the row.
			if ( vValue.inittype & VARTYPE_UNICODE ) {
				strwVal = vValue.lpwstr;
				vValue.lpwstr = (wchar_t *)strwVal.data(); // At it again!
			}
			else if ( vValue.inittype & VARTYPE_STRING ) {
				strVal = vValue.lpstr;
				vValue.lpstr = (char *)strVal.data(); // And again!
			}
			nResult = DllArrHandler(&vOut, dwSub1, 1,0,0,0,&vValue, 2);
			if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrHandler"));

			dwSub1++;
		}

		// Success at last!!
		nReturn = ReturnResult(&vOut);  // Return the map.
		
	}
	catch(CDllMapException &ex ) { // Dll Map services errors as excptions. Normally not shown to users.
		nExtError = ex.ResourceId();
	}
	catch(CDllGenericExecption &eg ) {      // Used for other programmer mistakes reported by the DLL.
		nExtError = IDS_ERROR_INTERNAL;      // 201038: Internal extender error
		DllStoreErrorInfo(eg.what());
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}
	catch(CUserException &eu ) { // User errors as excptions.
		nExtError = eu.ResourceId();
	}

	// Error?
	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	// Success.
	return nReturn;
}

/********************************************************************
 * Name: mpMapize
 *
 * Purpose: Converts a WIL array to a WIL asociative array.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "mpArrayize": parameter 1 =  WIL array.
 *
 *                        Return - WIL mapt made up of the input 
 *                                 array's first two column's 
 *                                 content. The content of each 
 *                                 row's column 0 becomes a map 
 *                                 key and each rows column 1 is
 *                                 the key's value.
 *
 * Notes:  Exceptions are reserved for unexpected, untoward events and
 *         not used for user input errors.
 ********************************************************************/
LONG CMapEx::mpMapize(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError   = 0;
	LONG nReturn     = 0;
	LONG nResult     = 0;
	DLLARRAYSERVICES1 ArrInfo = {0};
	DWORD_PTR dwSize = 0;
	vipervar vOut    = {0};
	DWORD dwDims[5]  = {0};

	try {

		// Check the calling DLL's feature level. WIL map support requires feature level 1011.
		if ( !IsInterfaceOk() ) 
			throw CUserException(IDS_ERROR_BADIFACE); // 201040: This extender function requires WIL DLL version 6.21bub or newer

		// It's an array that can be converted?
		nResult = DllArrayServices(lpvvArg, reinterpret_cast <LPBYTE> (&ArrInfo), sizeof(DLLARRAYSERVICES1), 1);
		if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrayServices"));

		if ( ArrInfo.cDims < 2 ) 
			throw CUserException(IDS_ERROR_ARREXPECTED); // 201039: Rank 2 or greater WIL array expected
		if ( ArrInfo.dwDim[0] < 1 || ArrInfo.dwDim[1] < 2 ) 
			throw CUserException(IDS_ERROR_ARREXPECTED); // 201039: Rank 2 or greater WIL array expected

		// Create a map.
		nResult = DllMapCreate( &vOut);
		if ( nResult ) throw CDllMapException(nResult);

		// Add the array's content to the map.
		DWORD dwRows = ArrInfo.dwDim[0];
		vipervar vKey = {0};
		vipervar vVal = {0};
		for ( DWORD dwRow = 0; dwRow < dwRows; dwRow++ ) {
			
			nResult = DllArrHandler(lpvvArg, dwRow, 0, 0, 0, 0, &vKey, 1);
			if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrHandler"));

			nResult = DllArrHandler(lpvvArg, dwRow, 1, 0, 0, 0, &vVal, 5);
			if ( nResult ) throw CDllGenericExecption(nResult, string("DllArrHandler"));

			// vKey is automaticaly converted to a string if possible.
			nResult = DllMapSet(&vOut, &vKey, &vVal);
			if ( nResult ) throw CDllMapException(nResult);

		}

		// Success at last!!
		nReturn = ReturnResult(&vOut);  // Return a map with the array's contents.
		
	}
	catch(CDllMapException &ex ) { // Dll Map services errors as excptions. Normally not shown to users.
		nExtError = ex.ResourceId();
	}
	catch(CDllGenericExecption &eg ) {      // Used for other programmer mistakes reported by the DLL.
		nExtError = IDS_ERROR_INTERNAL;      // 201038: Internal extender error
		DllStoreErrorInfo(eg.what());
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}
	catch(CUserException &eu ) { // User errors as excptions.
		nExtError = eu.ResourceId();
	}

	// Error?
	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	// The end...
	return nReturn;
}


///////////////////////////////////////////////////////////////////////////////////////
// Utility members.
///////////////////////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: IsMapParm
 *
 * Purpose: Checks whether or the passed in vipervar represents a
 *          WIL Map.
 *
 * Return: TRUE on success and FALSE on failure.   
 *
 ********************************************************************/
BOOL CMapEx::IsMap(LPVIPERVAR lpvvArg)
{
	if ( !(lpvvArg && lpvvArg->inittype == VARTYPE_ARRAY) )
		return FALSE;
	
	// Can only get a count when the vipervar holds a WIL map.
	DWORD_PTR dwDummy = 0;
	if ( DllMapCount(lpvvArg, &dwDummy) )
		return FALSE;

	return TRUE;
}

/********************************************************************
 * Name: StrToInt64 
 *
 * Purpose: Converts text to a 32-bit integer.
 *
 * Return: TRUE on success; otherwis, FALSE.  
 *
 * Notes: Throws errno value when out of range error or string pointer
 *        is NULL. Throws an EINVAL errno exception when the input does 
 *        not follow the pattern [whitespace] [sign] [base digits]]. 
 *
 *        Floating point numbers are not supported.
 *
 ********************************************************************/
BOOL CMapEx::StrToInt( LPTSTR pszNumber,   // In - String to convert
								 int nBase,        // In - Radix to convert from.
								 int * pnOut)      // Out - 32-bit integer number.
{
	CUserLocale Locale;

	if ( !pszNumber )
		return FALSE;
	
	// Clear trailing WIL whitespace which includes only 
	// the space and tab charaters.
	tstring Source = pszNumber;
   string::size_type Found = Source.find_last_not_of(_T(" \t"));
	if ( Found < (Source.length()-1) && (Found != string::npos) )
		Source.erase(Found+1);
	
	// Convert with error detection. 
	// (Errno is global so used with care.)
	errno_t saved_errno = errno;
	_set_errno(0);
	LPTSTR pEnd = NULL;
//#ifdef _M_AMD64
//	*pdwOut = _tcstoui64(Source.c_str(), &pEnd, nBase);
//#else
	*pnOut = _tcstol(Source.c_str(), &pEnd, nBase);
//#endif
	errno_t nErrno = errno;
   _set_errno(saved_errno);
	if ( nErrno == ERANGE || nErrno == EINVAL )
		return FALSE;

	// All characters must convert or its an error.
	if ( *pEnd != _T('\0') || Source.empty() ) 
		return FALSE;   

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////
// Required boilerplate extender interface members.
///////////////////////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: Identify (case 1 : WILExtenderQuery(W))
 *
 * Purpose: Case 1 is used to identify yourself to a caller.
 *          If lpvInfo is not NULL, it points to a character buffer where you may copy
 *          a string describing the extender.
 *          nCount is the size of the buffer pointed to by lpvInfo, in bytes.  When
 *          called by the WIL DLL, you can expect this to be a 256-byte buffer.
 *          The return value should be the version number of the extender.
 *
 * Return: Extender version number.
 *        
 * Notes: 
 *
 ********************************************************************/
LONG CMapEx::Identity( LPVOID lpvInfo,  // information buffer
                        LONG   nCount)   // additional data
{
   // Copy the description stored as a resource.
   if ( lpvInfo && nCount > 0 ) {
		LoadMyExtString(IDS_DESCRIPTION, m_szDescription, _countof(m_szDescription));  // From resource or language file.
      lstrcpyn((LPTSTR) lpvInfo, m_szDescription, min(nCount, _countof(m_szDescription)));
   }

	// Extract the dll's base name for use as the extender's name.
	TCHAR szFullName[MAX_PATH+1];
	int nResult = ::GetModuleFileName(CExtenderBase::m_hInstance, szFullName, _countof(szFullName));
	if ( nResult && nResult < _countof(szFullName) ) 
		_tsplitpath_s(szFullName, NULL, 0, NULL, 0, m_szExtName, _countof(m_szExtName), NULL, 0);

   // Attempt to extract version info from the dll's version resource block.
   DWORD dwExtVersion = 10001; // If all else fails.
   GetProductVersionNumber(m_hInstance, dwExtVersion);
   m_dwExtVersion = dwExtVersion;
   
   return (LONG) dwExtVersion;
}

/********************************************************************
 * Name: InterfaceVersion (case 2 : WILExtenderQuery(W))
 *
 * Purpose: case 2 is used to inform the WIL DLL of your preferred interface level.
 *          lpvInfo points to an array of LONG's, indicating the interface levels
 *          supported by the WIL DLL (each element represents one level).
 *          nCount indicates the number of elements in the array pointed to by lpvInfo.
 *       
 *          The return value must be the same as one of the elements in the lpvInfo
 *          array.  Currently, the WIL DLL supports interface levels 1 thu 5, so your
 *          return value should be 1 or 2.  Newly designed extenders (such as WILX)
 *          use interface level 2 and have a WilExtender2 entry point.  However, it
 *          is possible to design an extender to support multiple interface levels.
 *         
 *          If you return an invalid value, the extender will not be loaded and the
 *          AddExtender command will fail.           
 *
 * Return: Peferred interface version number
 *        
 * Notes:
 *
 ********************************************************************/
long CMapEx::InterfaceVersion( LPVOID lpvInfo,  // information buffer
                                LONG   nCount)   // additional data
{
   for (int i = 0; i < nCount; i++)
      if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL) // Found one we like.
         return INTERFACELEVEL;
      
   return -1;
}

/********************************************************************
 * Name: Initialization  (case -1 : WILExtender2(W))
 *
 * Purpose: Case -1 is called once, when your extender is loaded.  
 *          Use this member to perform your one-time initialization 
 *          tasks.  The return value should be the number of items 
 *          in your command map (functions and constants).  You may
 *          also return 0 if you are unable to determine the correct 
 *          number of items. You may return a negative value (-1) to 
 *          indicate that your extender was unable to successfully 
 *          initialize, for example, if it is running on an unsupported 
 *          platform or a dependent DLL could not be loaded.  This will
 *          cause the AddExtender command to fail.  You may additionally 
 *          provide a  string indicating the reason for the failure:
 *          lpvvArg points to a character buffer into which you may 
 *          copy an error message.
 *          "nArgCount" is the size of the buffer pointed to by lpvvArg, 
 *          in bytes. You can expect this to be a 256-byte buffer. 
 *          You can use this member implementation in your own extender 
 *          by changing the type name in used with the EntryT<> template 
 *          class referenced in this member.
 *
 * Return: > -1  = number of entries in command map.
 *         -1    = error condition.
 *
 * Notes:  While WIL is expecting the number of entries in your 
 *         command table this is a good place to add entries to the 
 *         table if you are creating your table at runtime instead of 
 *         at DLL load time like this example does.
 *        
 ********************************************************************/
LONG CMapEx::Initialization(HWND hWnd,           // hWnd of parent application.
                             HINSTANCE hInst,     // hInstance of parent application.
									  LPLONG lpLong,       // Pointer to your 32 bits.
									  LPVIPERVAR lpvvArg,  // Pointer to byte buffer.
									  LONG nArgCount)      // Size of lpvvArg in bytes.
{
	// You can also create extender table entries here.
	// Paramters are a pointer to a member, WIL function name, character length of name, 
	// required parameters, optional parameters, 0-16 paramter types.
	// For example,
	//m_ExtTable.push_back(EntryT<CMath64>(&CMath64::Add,_T("m64add"), 6, 2, 0, PTUni, PTUni)); 

	//return m_ExtTable.size();  // Number of extender implemented functions and constants.
	return _countof(m_ExtTable); // Number of extender implemented functions and constants.
}

/********************************************************************
 * Name: LoadTable  (case -2 : WILExtender2(W))
 *
 * Purpose: Case -2 is called repeatedly by the WIL interpreter.  On 
 *          each call, one CommandTable entry is passed back to the 
 *          WIL interpreter.  You can use this member implementation 
 *          in your own extender class by changing the type name in 
 *          used with the EntryT<> template class referenced in this
 *          member.
 *
 *
 * Return: 0 = done.
 *         1 = more entries.
 *
 * Notes:  
 *        
 ********************************************************************/
LONG CMapEx::LoadTable(HWND hWnd,           // hWnd of parent application.
                        HINSTANCE hInst,     // hInstance of parent application.
								LPLONG lpLong,       // Pointer to your 32 bits.
								LPVIPERVAR lpvvArg,  // Pointer to table entry structure.
								LONG nArgCount)      // Index of entry to process.
{    
   
   if ( nArgCount >= static_cast <LONG> (_countof(m_ExtTable)) )
		return 0; // Done.
	
	//if ( nArgCount >= static_cast <LONG> (m_ExtTable.size()) )
   //   return 0; // Done.
   
	LPTTABLESTRUCT lpTable = reinterpret_cast<LPTTABLESTRUCT>(lpvvArg);  // Convenient source code simplification.
   
	if ( lpTable ) {

		const EntryT<CMapEx> * pEntry = &(m_ExtTable[nArgCount]);

		// Make sure the name length is correct.
		assert(m_ExtTable[nArgCount].m_nSize == lstrlen(m_ExtTable[nArgCount].m_szName));

      lpTable->ident  = nArgCount;                    // Delimindex - must be consecutive!
		lpTable->type   = pEntry->m_nConstFunc;         // 2=FUNCTION 3=CONSTANT 4=COMMAND
      lpTable->param1 = pEntry->m_ArgCount.ParmCount; // (funct ? hi(reserved) lo(num-args))   (constant ? value)
      lpTable->ulMask = pEntry->m_ArgTypes.AllTypes;  // 64-bit mask (16 params, 4 bits each)
      lpTable->size   = pEntry->m_nSize;              // Length of function/constant name.

      // funct/const name - must be lowercase!
      ::lstrcpy(static_cast<LPTSTR>(lpTable->delim), pEntry->m_szName);
		::CharLower(static_cast<LPTSTR>(lpTable->delim));
   }
   
   return 1; // More entries.
}

/********************************************************************
 * Name: ShutDown  (case -3 : WILExtender2(W))
 *
 * Purpose: Case -3 is called during the WIL Shutdown process.  If 
 *          you need to do cleanup processing, add it to this member
 *          in your extender class.
 *
 * Return: 1 = success.
 *
 * Notes:  
 *        
 ********************************************************************/
LONG CMapEx::ShutDown(HWND hWnd,           // hWnd of parent application.
                       HINSTANCE hInst,     // hInstance of parent application.
                       LPLONG lpLong,       // Pointer to your 32 bits.
                       LPVIPERVAR lpvvArg,  // Pointer to parameter array.
                       LONG nArgCount)      // Parameter count for lpvvArg.
{    
   
   return 1;  // OK.
}
