#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

///////////////////////////////////////////////////////////////////////
//  Required preprocessor definitions.
///////////////////////////////////////////////////////////////////////
#define INTERFACELEVEL  5              // Extender interface level (DON'T CHANGE!)
#define EXTENDER_FEATURE_LEVEL 1011    // Required BATDLL feature interface level for maps. (Need WB2021B or newer.)
//#define EXTENDER_FEATURE_LEVEL 1010  // Switch to 1010 if the extender is not using WIL maps. This also allows the 
                                       // the extender to be used without map support on pre-2021B versions of WinBatch.
// WIL extender includes - do NOT remove.
#ifndef _WIN64
#pragma pack(push,4)
#endif
extern "C" {
#include "wwwpform.h"
#include "addons.h"
}
#ifndef _WIN64
#pragma pack(pop)
#endif

// Need for nothrow version of the std new operator
// used below so it is included here.
#include <new>  
/////////////////////////////////////////////////////
// Convenient linking to the version library.
#pragma comment(lib, "version.lib")

////////////////////////////////////////////////////////////////////////
// Stack buffer sizes.
///////////////////////////////////////////////////////////////////////
#define MAX_EXT_NAME MAX_PATH          // Max extender name size.
#define MAX_EXT_DESC MAX_PATH          // Max number of TCHARs in extender description.
#define MAX_ERROR_LEN 256              // Max number of TCHARs in extender reported errors.

//////////////////////////////////////////////////////////////////////
// Should be in addons.h.
//////////////////////////////////////////////////////////////////////
#ifdef UNICODE
#define LPTTABLESTRUCT LPTABLESTRUCTW
#else 
#define LPTTABLESTRUCT LPTABLESTRUCT
#endif

///////////////////////////////////////////////////////////////////////
// Extender template's abstact base class. 
class __declspec(novtable) CExtenderBase 
{
public :
	CExtenderBase() throw()
	{
		// Give the extern "C" entry points access to this.
		m_pImplementation = this;
	}
	~CExtenderBase() throw()
	{}

	// Can completely excape abstract class overhead.
	virtual LONG WILExtenderQuery(short  nRequest, LPVOID lpvInfo, LONG   nCount) = 0;
	virtual LONG WILExtender2(HWND hWnd, HINSTANCE hInst, LONG ViperInterfaceVersion, short CmdId, LPVIPERCALLBACK lpViperCallback, 
		                       LPVIPERSTRUCT lpViperStruct, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount, LPVIPERVAR lpViperReturn) = 0;
public:
	static HINSTANCE m_hInstance;
	static CExtenderBase * m_pImplementation;
};


///////////////////////////////////////////////////////////////////////////////////
// Extender Template - yet another example of CRTP.
// Use as the base class for WIL extenders.  
//
// Create an extender implementation class by using the following syntax:
// class CSpiffyExtender: public CWilExtenderT<CSpiffyExtender>
//{
//   Required members and extender WIL functions go here.
//};
template<class T>
class CWilExtenderT : public CExtenderBase
{
public:
	CWilExtenderT () throw() : m_ViperInterfaceVersion(0), m_lpViperCallback(NULL),
		m_lpViperStruct(NULL), m_lpViperReturn(NULL), m_dwExtVersion(0)
	{
		m_szExtName[0]     = _T('\0');
		m_szDescription[0] = _T('\0');
	}
	~CWilExtenderT () throw()
	{
	}

	//////////////////////////////////////////////////////////////////////////////
	// WIL request handlers redirect calls to the concrete extender implementation.
	
	LONG WILExtenderQuery( short  nRequest, LPVOID lpvInfo, LONG   nCount) {
		// Important!!!!!!!!!!
		// The following members must be implemented by the extender class derived from this template class.  
		switch ( nRequest ) {
			case 1: return static_cast<T*>(this)->Identity(lpvInfo, nCount);
			case 2: return static_cast<T*>(this)->InterfaceVersion( lpvInfo, nCount);
		}

		// Return 0 for any unknown requests, or any requests you don't handle.
		return 0; 
	}
	
	LONG WILExtender2(HWND hWnd, HINSTANCE hInst, LONG ViperInterfaceVersion, short CmdId, LPVIPERCALLBACK lpViperCallback, 
		               LPVIPERSTRUCT lpViperStruct, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount, LPVIPERVAR lpViperReturn)
	{
      // Update WIL provided property values for user by the class derived from this template class. 
		m_ViperInterfaceVersion = ViperInterfaceVersion;  
      m_lpViperCallback       = lpViperCallback;        
      m_lpViperStruct         = lpViperStruct;          
      m_lpViperReturn         = lpViperReturn; 
		
		// Important!!!!!!!!!!
		// The following members must be implemented by the extender class derived from this template class.  
		switch (CmdId) {
			case -1: return static_cast<T*>(this)->Initialization(hWnd, hInst, lpLong, lpvvArg, nArgCount);
			case -2: return static_cast<T*>(this)->LoadTable(hWnd, hInst, lpLong, lpvvArg, nArgCount);
			case -3: return static_cast<T*>(this)->ShutDown(hWnd, hInst, lpLong, lpvvArg, nArgCount);
			default: 
            
				assert(CmdId < _countof(m_ExtTable));

				// Call the extender's WIL function implementation member.
				return (static_cast<T*>(this)->*(m_ExtTable[CmdId].m_pMember))(hWnd, hInst, lpLong, lpvvArg, nArgCount);
      }

		return 0;  // Impossible to get here.
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	// Extender WIL function return handlers.
  
	inline int ReturnResult( LONG nNumber );
   inline int ReturnResult( int nNumber ) { return ReturnResult( (LONG) nNumber );};
   inline int ReturnResult( double dNumber );
   inline int ReturnResult( LPCTSTR pszStr );
#ifdef UNICODE
	inline int ReturnResult( LPCSTR pszStr );
#else
   inline int ReturnResult( LPCWSTR pszwStr );
#endif 
	inline int ReturnResult( LPVARIANT pVariant);
   inline int ReturnResult( LPVIPERVAR lpVv);
   int ReturnError( int nSeverity, int nErrorID);
   inline int ReturnCancel(){return -5;}

	////////////////////////////////////////////////////////////////////////////////////////////
	// WIL DLL extender services handlers.
	
	inline LONG_PTR DllServices(WORD wRequest, LPSTR szStr1, LPSTR szStr2, INT_PTR nData, DWORD_PTR dwBigData) throw()
   {
      return m_lpViperStruct->lpfnDllServices(m_lpViperStruct->hBatData,
                                              wRequest, 
                                              szStr1, 
                                              szStr2, 
                                              nData, 
                                              dwBigData);
   }

	inline LONG DllVarHandler(LPSTR lpVarName, LPVIPERVAR lpVipVar, LONG nRequest) throw()
   {
      return m_lpViperStruct->lpfnDllVarHandler(m_lpViperStruct->hBatData,lpVarName,lpVipVar,sizeof(vipervar),nRequest);
   }

   inline long DllArrHandler(LPVIPERVAR lpArr, DWORD dwSub1, DWORD dwSub2, DWORD dwSub3, DWORD dwSub4, DWORD dwSub5, LPVIPERVAR lpVipVar, LONG nRequest) throw()
   {
      return m_lpViperStruct->lpfnDllArrHandler(m_lpViperStruct->hBatData,
                                                lpArr,
                                                dwSub1,
                                                dwSub2,
                                                dwSub3,
                                                dwSub4,
                                                dwSub5,
                                                lpVipVar,
                                                sizeof(vipervar),
                                                nRequest);
   }

   inline long DllArrayServices(LPVIPERVAR lpArr, LPBYTE lpBuf, UINT uBufSize, LONG nRequest) throw()
   {
      return m_lpViperStruct->lpfnDllArrayServices(m_lpViperStruct->hBatData,lpArr,sizeof(vipervar),lpBuf,uBufSize,nRequest);
	}

	inline long DllArrayStringsToAnsi( LPVIPERVAR arr, UINT uVarSize) throw()
	{
		return m_lpViperStruct->lpfnDllArrayStringsToAns(m_lpViperStruct->hBatData,arr,uVarSize);
	}

	inline long DllArrayStringsToUnicode(LPVIPERVAR arr, UINT uVarSize) throw()
	{
		return m_lpViperStruct->lpfnDllArrayStringsToUnicode(m_lpViperStruct->hBatData,arr,uVarSize);
	}

	inline long DllStoreErrorInfo(LPCSTR lpErrorInfo) throw()
	{
		return m_lpViperStruct->lpfnDllStoreErrorInfo(m_lpViperStruct->hBatData,(LPSTR)lpErrorInfo);
	}

	inline LONG_PTR DllBinaryServices(LONG nBinBuf, LONG nData, LONG nRequest) throw()
	{
		return m_lpViperStruct->lpfnDllBinaryServices(m_lpViperStruct->hBatData,nBinBuf,nData,nRequest);
	}
	
	/////////////////////////////////////////////////
	// The Map related methods require compiling with the extender feature level (EXTENDER_FEATURE_LEVEL) defined to 1011 or higher.
#if EXTENDER_FEATURE_LEVEL >= 1011
	inline LONG DllMapServices(LPVIPERVAR pMap, UINT uMapSize, PDWORD_PTR pBuf, UINT uBufSize, LPVIPERVAR pKey, UINT uKeySize, LPVIPERVAR pValue, UINT uValSize, LONG nRequest) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,uMapSize,pBuf,uBufSize,pKey,uKeySize,pValue,uValSize,nRequest);
	}
	inline LONG DllMapCount(const LPVIPERVAR pMap, PDWORD_PTR pBuf) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),pBuf,sizeof(*pBuf),NULL,0,NULL,0,1);
	}
	inline LONG DllMapHashSize(PDWORD_PTR pHashSize) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,NULL,NULL,pHashSize,sizeof(*pHashSize),NULL,0,NULL,0,2);
	}
	inline LONG DllMapCreate(LPVIPERVAR pMap) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,NULL,0,NULL,0,3);
	}
	inline LONG DllMapGetANSI(const LPVIPERVAR pMap, LPVIPERVAR pKey, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,pKey,sizeof(*pKey),pValue,sizeof(*pValue),4);
	}
	inline LONG DllMapGetUnicode(const LPVIPERVAR pMap, LPVIPERVAR pKey, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,pKey,sizeof(*pKey),pValue,sizeof(*pValue),5);
	}
	inline LONG DllMapGetAsIs(const LPVIPERVAR pMap, LPVIPERVAR pKey, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,pKey,sizeof(*pKey),pValue,sizeof(*pValue),6);
	}
	inline LONG DllMapSet(const LPVIPERVAR pMap, const LPVIPERVAR pKey, const LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,pKey,sizeof(*pKey),pValue,sizeof(*pValue),7);
	}
	inline LONG DllMapFree(LPVIPERVAR pMap) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),NULL,0,NULL,0,NULL,0,8);
	}
	inline LONG DllMapGetIterator(const LPVIPERVAR pMap, PDWORD_PTR pIterator) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),pIterator,sizeof(*pIterator),NULL,0,NULL,0,9);
	}
	inline LONG DllMapGetNextANSI(const LPVIPERVAR pMap, LPVIPERVAR pKey, PDWORD_PTR pIterator, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),pIterator,sizeof(*pIterator),pKey,sizeof(*pKey),pValue,sizeof(*pValue),10);
	}
	inline LONG DllMapGetNextUnicode(LPVIPERVAR pMap, LPVIPERVAR pKey, PDWORD_PTR pIterator, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),pIterator,sizeof(*pIterator),pKey,sizeof(*pKey),pValue,sizeof(*pValue),11);
	}
	inline LONG DllMapGetNextAsIs(const LPVIPERVAR pMap, LPVIPERVAR pKey, PDWORD_PTR pIterator, LPVIPERVAR pValue) throw()
	{
		return m_lpViperStruct->lpfnDllMapServices(m_lpViperStruct->hBatData,pMap,sizeof(*pMap),pIterator,sizeof(*pIterator),pKey,sizeof(*pKey),pValue,sizeof(*pValue),12);
	}
#endif // EXTENDER_FEATURE_LEVEL >= 1011

	////////////////////////////////////////////////////////////////////////////////////////////
	// Utility methods
	BOOL GetProductVersionNumber(HMODULE hModule, DWORD& nVerson );
	int LoadMyExtString(int wID, LPTSTR lpBuffer,  int nBuffermax);      
	LPTSTR LastSlash(LPTSTR pszString);


	////////////////////////////////////////////////////////////////////////////////////////////
	// WIL extender function paramter types.
	enum ParamTypes : BYTE {
		PTInt = 0x0,  // Integer.
		PTStr = 0x1,  // Multi-byte (ANSI) string.
		PTFlt = 0x2,  // floating point.
		PTArr = 0x3,  // Array. 
		PTVar = 0x4,  // Variant. 
		PTUni = 0x5,  // Unicode string. 
		PTHyb = 0x6,  // hybrid string. 
		PTPDS = 0x7,  // platform-dependent string. 
		PTHPD = 0x8,  // hybrid platform-dependent string. 
		PTAny = 0xF,  // As-is (don't validate or convert.)
#ifdef UNICODE      
		PTTSt = PTUni // Unicode when compiled with Unicode Character Set.
#else
		PTTSt = PTStr // Multi-byte (ANSI) when compiled with Multi-byte Character Set.
#endif
	};

	////////////////////////////////////////////////////////////////////////////////////////////
	// WIL Extender table entry template.
	// If using a static table, this template can be used something like the following:
	//
	// CWilExtenderT<CSpiffyExtender>::EntryT<CSpiffyExtender> CWilExtenderT<CSpiffyExtender>::m_ExtTable[] =
	// {EntryT<CSpiffyExtender>(&CSpiffyExtender::Useful, _T("useful"), 7, 2, 0, PTStr, PTStr),
	// ... };
	////////////////////////////////////////////////////////////////////////////////////////////
	template<class T> 
	struct EntryT 
	{
	public:
		//////////////////////// Extender table entry constructors. 
		//
		// Function entry constructor.
		EntryT( LONG (T::*pMember) (HWND, HINSTANCE, LPLONG, LPVIPERVAR, LONG),
				 LPCTSTR szName, int nNameSize, BYTE nRequired, BYTE nOptional,
			    ParamTypes p1  = PTInt, ParamTypes p2  = PTInt, ParamTypes p3  = PTInt, ParamTypes p4  = PTInt, 
				 ParamTypes p5  = PTInt, ParamTypes p6  = PTInt, ParamTypes p7  = PTInt, ParamTypes p8  = PTInt,
				 ParamTypes p9  = PTInt, ParamTypes p10 = PTInt, ParamTypes p11 = PTInt,ParamTypes p12  = PTInt,
				 ParamTypes p13 = PTInt, ParamTypes p14 = PTInt, ParamTypes p15 = PTInt, ParamTypes p16 = PTInt) :
			m_pMember(pMember), m_szName(szName), m_nSize(nNameSize), m_ArgCount(nRequired, nOptional),
			m_ArgTypes(p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16),
			m_nConstFunc(FUNCTION)
		{}
		// Constant entry constuctor.
		EntryT( LPCTSTR szName, int nNameSize, int nValue) : m_pMember(NULL), 
			m_szName(szName), m_nSize(nNameSize), m_ArgCount(nValue),
			m_ArgTypes(0), m_nConstFunc(CONSTANT)
		{}

		// Extender table entry column values. 
		LONG (T::*m_pMember) (HWND, HINSTANCE, LPLONG, LPVIPERVAR, LONG); // Column 0 - Pointer to implementing member.
		LPCTSTR m_szName;         // Column 1 - Entry name.
		unsigned short  m_nSize;  // Column 2 - Number of characters in m_szName.
		int m_nConstFunc;         // Column 3 - FUNCTION or CONSTANT.
		union Parm_Cnt { 
			// Constructors.
			// - Function.
			Parm_Cnt(BYTE nRequired, BYTE nOptional): Required(nRequired), Optional(nOptional), notUsed(0)
			{}
			// - Constant.
			Parm_Cnt(int nValue) : ParmCount(nValue)
			{}
			struct {
				int Required:8; int Optional:8; int notUsed:16; 
			};
			int ParmCount;        
		} m_ArgCount;           // Column 4. Number of optional and required parameters or constant value.
		
		union Parm_Types { 
			// Constructors.
			// - Function.
			Parm_Types(	ParamTypes p1,  ParamTypes p2,  ParamTypes p3,  ParamTypes p4, 
				         ParamTypes p5,  ParamTypes p6,  ParamTypes p7,  ParamTypes p8,
							ParamTypes p9,  ParamTypes p10, ParamTypes p11, ParamTypes p12, 
							ParamTypes p13, ParamTypes p14, ParamTypes p15, ParamTypes p16) :
			pt1(p1), pt2(p2), pt3(p3), pt4(p4), pt5(p5), pt6(p6), pt7(p7), pt8(p8), 
			pt9(p9), pt10(p9), pt11(p11), pt12(p12), pt13(p13), pt14(p14), pt15(p15), pt16(p16)
			{}
			// - Constant.
			Parm_Types(__int64 notUsed): AllTypes(notUsed)
			{}

			struct {
				__int64 pt1:4;__int64 pt2:4;__int64 pt3:4;__int64 pt4:4;__int64 pt5:4;__int64 pt6:4;__int64 pt7:4;__int64 pt8:4;
            __int64 pt9:4;__int64 pt10:4;__int64 pt11:4;__int64 pt12:4;__int64 pt13:4;__int64 pt14:4;__int64 pt15:4;__int64 pt16:4;
			} ;
         __int64 AllTypes;
		} m_ArgTypes;           // Column 5 - Paramter type bit mask.
	};
 
protected:

	///////////////////////////////////////////////////////////////////////////////////////////
	// Members representing the pointers and values passed to the extender form the WIL dll on 
	// every call made by the WIL DLL to the extender's WILExtender2 export.  These are stored
	// as class properties to give class methods easy access to them.
	// While they are unlikely to change, they are refreshed each time the WILExtender2 
	// function is called in the unlikely event that they do change.
	///////////////////////////////////////////////////////////////////////////////////////////
	LONG            m_ViperInterfaceVersion;       // WIL interface version.
   LPVIPERCALLBACK m_lpViperCallback;             // Pointer to return value callback function.
   LPVIPERSTRUCT   m_lpViperStruct;               // Pointer to information structure.
   LPVIPERVAR      m_lpViperReturn;               // Pointer to return value structure.
   
	// Convenient data members.
	DWORD           m_dwExtVersion;                // Extender version.
	TCHAR           m_szDescription[MAX_EXT_DESC]; // Extender description string.
   TCHAR           m_szExtName[MAX_EXT_NAME];     // Name of extender. (Must be set for langauge files.)

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Could use either the STL:vector or built-in arrays for the extender's table.

	//vector<EntryT<T>> m_ExtTable;                  // The one and only extender functions and constants table.
public:
	static const EntryT<T> m_ExtTable[];            // The one and only extender functions and constants table.
private:
	// Private data members.
};

//////////////////////////////////////////////////////////////////////////////////////
// Extender WIL function return handlers

/********************************************************************
 * Name: ReturnResult - long integer
 *
 * Purpose: Passes long integer result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
inline int CWilExtenderT<T>::ReturnResult( LONG lNumber ) // In - Long integer.
{
   m_lpViperReturn->x        = lNumber;
   m_lpViperReturn->inittype = VARTYPE_INT;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - integer
 *
 * Purpose: Passes integer result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  Inlined in CWilExtender.h
 *        
 ********************************************************************/
//template<class T> 
//int CWilExtenderT<T>::ReturnResult( int nNumber ) // In - integer.

/********************************************************************
 * Name: ReturnResult - double
 *
 * Purpose: Passes double result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(double dNumber) // In - floating point number.
{
   m_lpViperReturn->flt      = dNumber;
   m_lpViperReturn->inittype = VARTYPE_FLOATNUM;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - string
 *
 * Purpose: Passes string result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(LPCTSTR pszStr) // In - pointer to string
{
#ifdef UNICODE
      m_lpViperReturn->lpwstr   = (LPTSTR) pszStr;
      m_lpViperReturn->inittype = VARTYPE_UNICODE;
#else
      m_lpViperReturn->lpstr    = (LPTSTR) pszStr;                      
      m_lpViperReturn->inittype = VARTYPE_STRING;
#endif

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}
#ifdef UNICODE
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(LPCSTR pszStr) // In - pointer to  ANSI string
{
	m_lpViperReturn->lpstr    = (LPSTR) pszStr;                      
	m_lpViperReturn->inittype = VARTYPE_STRING;

	return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}
#else
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(LPCWSTR pszwStr) // In - pointer to Unicode string
{
	m_lpViperReturn->lpwstr   = (LPWSTR) pszwStr;
	m_lpViperReturn->inittype = VARTYPE_UNICODE;

	return m_lpViperCallback(RESULT_SUCCESS, 
								    m_ViperInterfaceVersion,                     
								    m_lpViperStruct,                             
								    m_lpViperReturn);
}
#endif


/********************************************************************
 * Name: ReturnResult - variant
 *
 * Purpose: Passes variant result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(LPVARIANT pVariant) // In - pointer to variant
{
   m_lpViperReturn->inittype = VARTYPE_VARIANT;
   m_lpViperReturn->variant  = *pVariant;   // Note: shallow copy.

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - vipervar
 *
 * Purpose: Passes vipervar result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
inline int CWilExtenderT<T>::ReturnResult(LPVIPERVAR lpVv) // In - pointer to vipervar
{
	*m_lpViperReturn = *lpVv;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnError 
 *
 * Purpose: Passes severity code and error message to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes: nSeverity -  MUST be -1, -2 or -3.
 *                     Use -1 for most run of the mild errors.
 *                         -2 for serious errors
 *                         -3 for absolutely fatal "we better die" errors
 *
 *        nErrorID - is the string ID from the RC file that contains 
 *                   the error message.  The first three characters of
 *                   the string should be the same as the string ID, 
 *                   and the fourth character should be a colon.
 *        
 ********************************************************************/
template<class T> 
int CWilExtenderT<T>::ReturnError( int nSeverity, // In- severity code.
                                   int nErrorID)  // In- resource id.
{
	TCHAR sztError[MAX_ERROR_LEN];    // Note: max error length.
   
	// Load the error message.
   LoadMyExtString( nErrorID, sztError, MAX_ERROR_LEN);
 
#ifdef UNICODE 
	CHAR szaError[MAX_ERROR_LEN];
   // Unicode error messages must be converted to ANSI
   // because of WinBatch cannot handle Unicode error message.
	// Using system code page which may or may not be different than the user's current code page.
	if (::WideCharToMultiByte( CP_ACP, 0, sztError, -1, szaError, MAX_ERROR_LEN, NULL, NULL ) == 0) 
		::lstrcpynA(szaError, "Text conversion error", MAX_ERROR_LEN);
	m_lpViperReturn->lpstr = (LPSTR) szaError;
#else
	m_lpViperReturn->lpstr = (LPSTR) sztError;
#endif

   m_lpViperReturn->inittype = VARTYPE_STRING;

   return m_lpViperCallback(nSeverity, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

//////////////////////////////////////////////////////////////////////////////////////
// Template utility methods

/********************************************************************
 * Name: GetProductVersionNumber
 *
 * Purpose:  This member extracts the product verion number from
 *           the dll's version resource.
 *
 * Return: TRUE if successful.
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
BOOL CWilExtenderT<T>::GetProductVersionNumber(HMODULE hModule, //In-handle to this dll. 
                                               DWORD& nVersion ) //In/out - ref. dword/version number.
{
	TCHAR filename[MAX_PATH];
	BYTE * pVersionInfo = NULL;
	BOOL bReturn        = FALSE;

	// Get module file name.
	DWORD dwLen = ::GetModuleFileName(hModule, filename,_countof(filename));
	if ( dwLen <= 0 ) 
		goto LabelError;

	// Read the raw file version info.
	DWORD dwDummyHandle; // Will always be set to zero.
	dwLen = ::GetFileVersionInfoSize(filename, &dwDummyHandle);
	if ( dwLen <= 0 ) 
		goto LabelError;

	pVersionInfo = (BYTE *) new(std::nothrow)(BYTE[dwLen]);
	if ( pVersionInfo == NULL )
		goto LabelError;

	if (!::GetFileVersionInfo(filename, 0, dwLen, pVersionInfo))
		goto LabelError;

	// Get translation info.
	struct TRANSLATION {
		WORD langID;        // language ID
		WORD charset;       // character set (code page)
	} translation;         // Language and code page of resource.
	translation.langID  = 1033;  // Chauvinistically, US English.
	translation.charset = 1200;  // Unicode.

	LPVOID lpvi;
	UINT   iLen;
	if ( ::VerQueryValue(pVersionInfo, _T("\\VarFileInfo\\Translation"), &lpvi, &iLen) && iLen >= 4 ) 
		translation = *(TRANSLATION*)lpvi;

	// Get the version number string. WIL likes to user the string version.
	TCHAR szTemp[128];
	::wsprintf(szTemp, _T("\\StringFileInfo\\%04x%04x\\%s"), translation.langID, translation.charset, _T("FileVersion"));
	iLen = 0;
	LPTSTR pVal = NULL;
	if ( !::VerQueryValue(pVersionInfo, szTemp, (LPVOID*)&pVal, &iLen) ) 
		goto LabelError;

	// Errno is global so used with care.
	errno_t saved_errno = errno;
	_set_errno(0);
	DWORD dwVersion = _tstol(pVal);
	errno_t nErrno = errno;
   _set_errno(saved_errno);
	if ( nErrno == ERANGE || nErrno == EINVAL ) 
		goto LabelError;

	nVersion = dwVersion;
	bReturn = TRUE;  // Success.

LabelError:

	if ( pVersionInfo )
		delete [] pVersionInfo;

	return bReturn;
}

/********************************************************************
 * Name: LoadMyExtString
 *
 * Purpose: Loads string from resource or language file.  
 *
 * Return: Number of chars in output.
 *
 * Notes:  Ugly bit of code.
 *        
 ********************************************************************/
template<class T> 
int CWilExtenderT<T>::LoadMyExtString(int wID,                //In - string resource id.
                                      LPTSTR lpBuffer,        //In/out - allocated memory/string.
                                      int nBuffermax)         //In - size of allocated memory.
{
   LPTSTR       lpSlash;
   int          nCharsBack, i;
	static TCHAR szLangIni[MAX_PATH] = { _T('\0') };
   static int  nIniExist        = 0;           // 0=uninit   1=notrans   2=xlate it
   static TCHAR szLangSection[MAX_PATH] = { _T('\0') };
   static TCHAR szLANGDLL[]     = _T("WWWELANG");
   static TCHAR szXXX[]         = _T("XXX");
   static TCHAR szMAIN[]        = _T("MAIN");
   static TCHAR szINTL[]        = _T("INTL");
   static TCHAR szLibIni[]      = _T("WWWBATCH.INI");
   static TCHAR szPanic[256];
   TCHAR        SafeBuffer[20];
   TCHAR        szKey[20];
   TCHAR        szFullName[MAX_PATH];
   
   // If we errored allocating a buffer, we will use our panic buffer.
   if ( !nBuffermax ) {
      lpBuffer = szPanic;
      nBuffermax = _countof(szPanic);
   }
   
   if (nIniExist == 0 || wID == 0)  {    // Process not initialized
   
      nIniExist    = 1;
      szLangIni[0] = '\0';
      
      // First see if a language override has been installed
      ::GetPrivateProfileString(szMAIN, _T("Language"), szXXX, SafeBuffer, _countof(SafeBuffer), szLibIni);
      
      if (lstrcmp(SafeBuffer, szXXX) == 0)  // no language override
         ::GetProfileString(szINTL, _T("sLanguage"), _T("enu"), SafeBuffer, _countof(SafeBuffer));  // look foreign lang.  Def to enu
      
      // Get DLL location for search for our lng dll
      i = ::GetModuleFileName(CExtenderBase::m_hInstance, szLangIni, _countof(szLangIni));
      
      if (i == 0)
         szLangIni[0] = _T('\0');
      else {
         lpSlash = LastSlash(szLangIni);
         
         if (lpSlash)
				*(::CharNext(lpSlash)) = _T('\0');
         else
            szLangIni[0] = _T('\0');
      }
      
		::lstrcat(szLangIni, szLANGDLL);     // glue on WWWELANG
		::lstrcat(szLangIni, _T("."));       // glue on .
		::lstrcat(szLangIni, SafeBuffer);    // glue on enu or whatnot
      
      LPTSTR lpDummy = NULL;
		DWORD dwLen    = ::SearchPath(NULL,         // pointer to search path
                                    szLangIni,    // pointer to file name
                                    NULL,         // pointer to extension
                                    MAX_PATH,     // size, in characters, of return buffer
                                    szFullName,   // pointer to buffer for found file name
                                    &lpDummy);    // pointer to pointer to file component
      if ( dwLen == 0 ) {
         wsprintf((LPTSTR) szLangIni, _T("%s.%s"), (LPTSTR) szLANGDLL, (LPTSTR) SafeBuffer);
			dwLen    = ::SearchPath(NULL,         // pointer to search path
                                 szLangIni,    // pointer to file name
                                 NULL,         // pointer to extension
                                 MAX_PATH,     // size, in characters, of return buffer
                                 szFullName,   // pointer to buffer for found file name
                                 &lpDummy);    // pointer to pointer to file component
         if (dwLen)
				::lstrcpyn(szLangIni, szFullName, _countof(szLangIni));
      }
      
      if (dwLen) {

         // Create the extenders section name from the extender name.
			::lstrcpyn(szLangSection, m_szExtName, _countof(szLangSection)-1);
			::lstrcpyn(szLangSection, _T(" Extender"), MAX_EXT_NAME);
         nIniExist = 2;
      }
   }
   
   if ( (wID >= 100) && (nIniExist == 2) ) {
      szKey[0] = _T('X');
      szKey[1] = _T('\0');
      _ltot_s(wID, szKey + 1, _countof(szKey), 10);
      

      nCharsBack = ::GetPrivateProfileString(szLangSection, szKey, szXXX, lpBuffer, nBuffermax, szLangIni);
      
      if ( lstrcmp(lpBuffer, szXXX) == 0 ) {
         lpBuffer[0] = _T('\0');
         nCharsBack = ::LoadString(CExtenderBase::m_hInstance, wID, lpBuffer, nBuffermax);
      }
   }
   else {
      lpBuffer[0] = _T('\0');
      nCharsBack = ::LoadString(CExtenderBase::m_hInstance, wID, lpBuffer, nBuffermax);
   }
      
   while (*lpBuffer) {
      if (*lpBuffer == _T('~'))
         *lpBuffer = _T('\n');
      
      if (*lpBuffer == _T('|'))
         *lpBuffer = _T('\t');
      
      lpBuffer = ::CharNext(lpBuffer);
   }

   return nCharsBack;
}

/********************************************************************
 * Name: LastSlash
 *
 * Purpose: Returns a pointer to the last slash (forward or 
 *          backward slash) found in a string. Supports MBCS or
 *          Unicode compiled DLLs.
 *
 * Return: Pointer to the last slash or NULL if none found.
 *
 * Notes:  
 *        
 ********************************************************************/
template<class T> 
LPTSTR CWilExtenderT<T>::LastSlash(LPTSTR pszString)
{
  LPTSTR pszPtr   = pszString;
  LPTSTR pszSlash = NULL;

  for ( ;*pszPtr; pszPtr = ::CharNext(pszPtr))
    if ((*pszPtr == _T('\\')) || (*pszPtr == _T('/')))
      pszSlash = pszPtr;

  return pszSlash;
}

///////////////////////////////////////////////////////////////////////////

