// Addons.h   Addons file for WIL extenders
//
// This can be used to build version 2/3, version 4, or version 5 extenders
//
// INTERFACELEVEL must be defined before including this header

#ifndef _ADDONS_H_
#define _ADDONS_H_

// Need 64-bit DWORD_PTR for 64-bit versions of extenders so
// assume a 32-bit compiler and define it to DWORD when the 
// C/C++ compiler does not have architecture Marcos pre-defined.
// ANSI C/C++ complian compilers will not generate a compiler error
// for redeclaring with typedef as long as the type-declaration 
// matches the previously existing type-declaration for the synonym. 
#ifndef _M_X64         
typedef DWORD DWORD_PTR;
typedef LONG  LONG_PTR;
#endif  // !_M_X64

#define DllExport __declspec(dllexport)

typedef LPVOID LPBDS;                  // not really
typedef LPVOID LPRESULTVAR;            // not really
typedef LONG (FARPASCAL * DLLSERVICES) (HGLOBAL, WORD, LPSTR, LPSTR, int, DWORD_PTR);

typedef int FAR PASCAL AddOnCallback (int, LONG, LPSTR, BOOL, LONG, LPSTR);
typedef AddOnCallback FAR *lpAddOnCallback;

#define VARTYPE_UNDEF         0        // These are bitmapped, var can be both
#define VARTYPE_INT           1        // Numeric and string, must check bits
#define VARTYPE_STRING        2        // ANSI string
#define VARTYPE_FILE          5        // FILE and INT bits set
#define VARTYPE_FLOATNUM     32        // floating point
#define VARTYPE_BINARY       65        // BINARY and INT bits set
#define VARTYPE_UNICODE     128        // a Unicode (wide) string
#define VARTYPE_ARRAY       256        // an array
#define VARTYPE_VARIANT     512        // a COM (OLE) variant

typedef struct vipervariable2          // version 2/3 variable
{
  double  flt;                         // 8-byte floating point number
  LONG    x;                           // if int variable, value is here
  int     inittype;                    // bitmask of VARTYPE_* bits
  union
  {
    LPSTR     lpstr;                   // pointer to string
    DWORD_PTR handle;                  // handle for miscellaneous object types
  };
} vipervar2, FAR *LPVIPERVAR2;

typedef struct vipervariable4          // version 4 variable
{
  VARIANT variant;                     // if Variant, variable here.
  double  flt;                         // 8-byte floating point number
  LONG    x;                           // if int variable, value is here
  int     inittype;                    // bitmask of VARTYPE_* bits
  union
  {
    LPSTR     lpstr;                   // pointer to string
    DWORD_PTR handle;                  // handle for miscellaneous object types
  };
  DWORD   reserved;                    // reserved for future expansion
} vipervar4, FAR *LPVIPERVAR4;

typedef struct vipervariable5          // version 5 variable
{
  VARIANT variant;                     // if Variant, variable here.
  union
  {
    double  flt;                       // 8-byte floating point number
    INT64   qword;                     // 64-bit integer value
  };
  LONG    x;                           // if int variable, value is here
  int     inittype;                    // bitmask of VARTYPE_* bits
  union
  {
    LPSTR  lpstr;                      // pointer to string
    LPWSTR lpwstr;                     // pointer to Unicode (wide) string
    DWORD_PTR handle;                  // handle for miscellaneous object types
  };
  DWORD   reserved;                    // reserved for future expansion
} vipervar5, FAR *LPVIPERVAR5;

#if INTERFACELEVEL == 5
  #define vipervar vipervar5
  #define LPVIPERVAR LPVIPERVAR5
#elif INTERFACELEVEL == 4
  #define vipervar vipervar4
  #define LPVIPERVAR LPVIPERVAR4
#elif INTERFACELEVEL >= 2
  #define vipervar vipervar2
  #define LPVIPERVAR LPVIPERVAR2
#else
  #error INTERFACELEVEL is invalid or undefined
#endif

/*----------------------------------------------------------------------------
LONG FARPASCAL DllVarHandler(HGLOBAL hBatData, LPSTR lpVarName, LPVIPERVAR vv, UINT uVarSize, LONG nRequest)
LONG FARPASCAL DllVarHandlerW(HGLOBAL hBatData, LPWSTR lpwVarName, LPVIPERVAR vv, UINT uVarSize, LONG nRequest)

  hBatData    =  lpViperStruct->hBatData
  lpVarName   =  name of WIL variable (ANSI version)
  lpwVarName  =  name of WIL variable (Unicode version)
  vv          =  vipervar for input or output, depending on request
  uVarSize    =  sizeof(vipervar)
  nRequest    =  operation to perform:
    1  =  get value (convert strings to ANSI)     (vv = output value)
    2  =  set (and clear) value                   (vv = value to set)
    3  =  clear value                             (vv = 0)
    4  =  get value (convert strings to Unicode)  (vv = output value)
    5  =  get value (ANSI or Unicode strings)     (vv = output value)

    Note that a WIL variable can be both ANSI and Unicode, but a vipervar cannot.

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  Invalid lpVarName
     3  lpVarName too long
     4  vv buffer too small
     5  GlobalLock(hBatData) failed
     6  GlobalLock(hVarHandle) failed
     7  Specified variable name not found
     8  Error getting variant value
     9  Variable table is full
    10  Error setting variant value
    11  Error allocating memory for string
    12  Invalid request
    13  Invalid variable "pointer"
    14  Unable to access variable via pointer
    15  Error accessing variable table
    16  vv cannot point to string in WIL string table
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLVARHANDLER) (HGLOBAL, LPSTR, LPVIPERVAR, UINT, LONG);
typedef LONG (FARPASCAL * DLLVARHANDLERW) (HGLOBAL, LPWSTR, LPVIPERVAR, UINT, LONG);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllArrHandler(HGLOBAL hBatData, LPVIPERVAR arr, DWORD dwSub1, DWORD dwSub2, DWORD dwSub3, DWORD dwSub4, DWORD dwSub5, LPVIPERVAR vv, UINT uVarSize, LONG nRequest)

  hBatData  =  lpViperStruct->hBatData
  arr       =  vipervar identifying array
  dwSub1 - dwSub5  =  subscripts identifying array element
  vv        =  vipervar for input or output, depending on request
  uVarSize  =  sizeof(vipervar)
  nRequest  =  operation to perform:
    1  =  get value (convert strings to ANSI)     (vv = output value)
    2  =  set (and clear) value                   (vv = value to set)
    3  =  clear value                             (vv = 0)
    4  =  get value (convert strings to Unicode)  (vv = output value)
    5  =  get value (ANSI or Unicode strings)     (vv = output value)

    Note that an array variable can be both ANSI and Unicode, but a vipervar cannot.

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  arr buffer too small
     3  arr variable is not an array
     4  Array is empty or invalid
     5  Array data is invalid
     6  vv buffer too small
     7  GlobalLock(hBatData) failed
     8  Subscript out of bounds
     9  Error accessing array data
    10  Error reading or writing array data
    11  GlobalLock(hVarHandle) failed
    12  Error getting variant value
    13  Cannot create an embedded array
    14  Error setting variant value
    15  Error allocating memory for string
    16  Invalid request
    17  vv cannot point to string in WIL string table
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLARRHANDLER) (HGLOBAL, LPVIPERVAR, DWORD, DWORD, DWORD, DWORD, DWORD, LPVIPERVAR, UINT, LONG);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllArrayServices(HGLOBAL hBatData, LPVIPERVAR arr, UINT uVarSize, LPBYTE lpBuf, UINT uBufSize, LONG nRequest)

  hBatData  =  lpViperStruct->hBatData
  arr       =  vipervar identifying array
  uVarSize  =  sizeof(vipervar)
  lpBuf     =  buffer for input or output, depending on request (cast to LPBYTE)
  uBufSize  =  size of buffer pointed to by lpBuf
  nRequest  =  operation to perform:
    1  =  get array info    (lpBuf = pointer to DLLARRAYSERVICES1 structure)
    2  =  create array      (lpBuf = pointer to DWORD[5] containing array dimensions
    3  =  initialize array  (lpBuf = vipervar containing value to set)
    4  =  free array        (lpBuf = 0)

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  arr buffer too small
     3  arr variable is not an array
     4  Array is empty or invalid
     5  GlobalLock(hBatData) failed
     6  GlobalLock(hVarHandle) failed
     7  lpBuf buffer too small
     8  Create array failed: Invalid dimension
     9  Create array failed: Cannot have 0 dimension followed by non-0 dimension
    10  Create array failed: Array too large
    11  Create array failed: Unable to allocate or lock memory
    12  Initialize array failed: Value is invalid type
    13  Initialize array failed: Array is empty
    14  Initialize array failed: Error accessing array data
    15  Initialize array failed: Out of memory for strings
    16  Invalid request
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLARRAYSERVICES) (HGLOBAL, LPVIPERVAR, UINT, LPBYTE, UINT, LONG);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllArrayStringsToAnsi(HGLOBAL hBatData, LPVIPERVAR arr, UINT uVarSize)

  hBatData  =  lpViperStruct->hBatData
  arr       =  vipervar identifying array
  uVarSize  =  sizeof(vipervar)

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  arr buffer too small
     3  arr variable is not an array
     4  Array is empty or invalid
     5  GlobalLock(hBatData) failed
     6  GlobalLock(hVarHandle) failed
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLARRAYSTRINGSTOANSI) (HGLOBAL, LPVIPERVAR, UINT);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllArrayStringsToUnicode(HGLOBAL hBatData, LPVIPERVAR arr, UINT uVarSize)

  hBatData  =  lpViperStruct->hBatData
  arr       =  vipervar identifying array
  uVarSize  =  sizeof(vipervar)

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  arr buffer too small
     3  arr variable is not an array
     4  Array is empty or invalid
     5  GlobalLock(hBatData) failed
     6  GlobalLock(hVarHandle) failed
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLARRAYSTRINGSTOUNICODE) (HGLOBAL, LPVIPERVAR, UINT);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllArraySwapElements(HGLOBAL hBatData, LPVIPERVAR arr, UINT uVarSize, DWORD dwSubA1, DWORD dwSubA2, DWORD dwSubA3, DWORD dwSubA4, DWORD dwSubA5, DWORD dwSubB1, DWORD dwSubB2, DWORD dwSubB3, DWORD dwSubB4, DWORD dwSubB5)

  hBatData  =  lpViperStruct->hBatData
  arr       =  vipervar identifying array
  uVarSize  =  sizeof(vipervar)
  dwSubA1 - dwSubA5  =  subscripts identifying array element A
  dwSubB1 - dwSubB5  =  subscripts identifying array element B

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  arr buffer too small
     3  arr variable is not an array
     4  Array is empty or invalid
     5  Array data is invalid
     6  GlobalLock(hBatData) failed
     7  Subscript out of bounds
     8  Error accessing array data
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLARRAYSWAPELEMENTS) (HGLOBAL, LPVIPERVAR, UINT, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD, DWORD);

/*---------------------------------------------------------------------------
LONG_PTR FARPASCAL DllBinaryServices(HGLOBAL hBatData, LONG nBinBuf, LONG nData, LONG nRequest)

  hBatData  =  lpViperStruct->hBatData
  nBinBuf   =  binary buffer handle
  nData     =  additional data, depending on request
  nRequest  =  operation to perform:
    1  =  allocate buffer      (nData = buffer size)  (return = buffer handle)       (nBinBuf = 0)
    2  =  deallocate buffer    (nData = 0)            (return = 0)
    3  =  get pointer to data  (nData = 0)            (return = LPSTR cast to LONG_PTR)
    4  =  get binary EOD       (nData = 0)            (return = binary EOD)
    5  =  get buffer size      (nData = 0)            (return = buffer size)
    6  =  set binary EOD       (nData = new EOD)      (return = previous EOD)

  Returns a non-negative value on success, or a negative value on error:
     -1  Function failed
     -2  Invalid binary buffer handle
     -3  GlobalLock(hBatData) failed
     -4  GlobalLock(hVarHandle) failed
     -5  Too many open binary buffers
     -6  Unable to allocate or lock binary buffer
     -7  Specified EOD beyond end of buffer
     -8  Invalid request
-----------------------------------------------------------------------------*/

typedef LONG_PTR (FARPASCAL * DLLBINARYSERVICES) (HGLOBAL, LONG, LONG, LONG);

/*---------------------------------------------------------------------------
LONG FARPASCAL DllStoreErrorInfo(HGLOBAL hBatData, LPSTR lpErrorInfo)

  hBatData     =  lpViperStruct->hBatData
  lpErrorInfo  =  error information string to store

  Returns 0 on success, or a positive value on error:
     1  Function failed
     2  GlobalLock(hBatData) failed
-----------------------------------------------------------------------------*/

typedef LONG (FARPASCAL * DLLSTOREERRORINFO) (HGLOBAL, LPSTR);


/*---------------------------------------------------------------------------
 * LONG FARPASCAL DllMapServices(HGLOBAL hBatData,LPVIPERVAR pVvMap,UINT uMapSize,PDWORD_PTR pBuffer,UINT uBufSize,LPVIPERVAR pVvKey,UINT uKeySize,LPVIPERVAR pVvValue, UINT uValSize,LONG nRequest)
 *
 *  Purpose: Provide map functionality to WIL extenders.
 *
 *  Return:
 *   0  =  Operations successfully performed.
 *   1  =  Function failed.
 *   2  =  Map variable too small.
 *   3  =  Map variable is not a map.
 *   4  =  Map variable is invalid.
 *   5  =  Map data is invalid.
 *   6  =  GlobalLock(hBatData) failed.
 *   7  =  pBuffer to small or can't read/write too.
 *   8  =  Error obtaining variant value 
 *   10 =  Error reading/writing map data.
 *	  11 =  Map variable contains data (memory must be initialized to zero)
 *   12 =  Unable to allocate or lock memory.
 *   13 =  pVvKey to small or can't read/write too.
 *   14 =  bad key value.
 *   15 =  Error allocating memory for string.
 *   16 =  Vipervar not a string.
 *   17 =  pVvValue Vipervar to small.
 *   18 =  Error reading/writing Value vipervar.
 *   19 =  pVvValue is not a valid type.
 *   20 =  Error setting variant value
 *   21 =  Hash table size too big or too small
 *   -1 =  key not found (request #s 4,5,6,7.) pBuffer unchanged.
 *
 *  nRequest possible values and their meaning:   
 *   1  =  # of pairs in map                       (pVvMap = map, sizeof vipervar, pBuffer = # key/value pairs, uBufSize = sizeof DWORD_PTR) 
 *   2  =  set hash table size                     (pBuffer = in new size/out=previous setting, uBufSize = sizeof DWORD_PTR)
 *   3  =  create Map                              (pVvMap = map, uMapSize = sizeof vipervar)
 *   4  =  get value (convert strings to ANSI)     (pVvMap = map, uMapSize = sizeof vipervar, pVvKey = in Key, pVvValue = outputvalue)
 *   5  =  get value (convert strings to Unicode)  (pVvMap = map, uMapSize = sizeof vipervar, pVvKey = in Key, pVvValue = outputvalue)
 *   6  =  get value (ANSI or Unicode strings)     (pVvMap = map, uMapSize = sizeof vipervar, pVvKey = in Key, uKeySize = sizeof vipervar, 
 *                                                  pVvValue = outputvalue, uValSize = sizeof vipervar)
 *   7  =  set key/value pair or change a value    (pVvMap = map, uMapSize = sizeof vipervar, pVvKey = in Key, pVvValue = outputvalue)
 *   8  =  Free a map removing all references.     (pVvMap = map, uMapSize = sizeof vipervar)
 *   9  =  get iterator.                           (pVvMap = map, uMapSize = sizeof vipervar, pBuffer = &DWORD_PTR out = iterator, uBufSize = sizeof DWORD_PTR )
 *  10  =  Next key/value pair (convert value strings to ANSI)    (pVvMap = map, uMapSize = sizeof vipervar, pBuffer = iterator, pVvKey = out Key, 
 *                                                                 uKeySize = sizeof vipervar, pVvValue = out value, uValSize = sizeof vipervar)
 *  11  =  Next key/value pair (convert value strings to Unicode) (pVvMap = map, uMapSize = sizeof vipervar, pBuffer = iterator, pVvKey = out Key, 
 *                                                                 uKeySize = sizeof vipervar, pVvValue = out value, uValSize = sizeof vipervar)
 *  12  =  Next key/value pair (leave value strings as is)        (pVvMap = map, uMapSize = sizeof vipervar, pBuffer = iterator, pVvKey = in Key,
 *                                                                 uKeySize = sizeof vipervar, pVvValue = out value, uValSize = sizeof vipervar)
 *  10,11,12 pBuffer = PDWORD_PTR first obtain using request coded 9. *pBuffer set to 0 when last pair returned.
 *  10,11,12 pVvkey keys are always returned as an ANSI string (->lpstr.)
 ****************************************************************************/
typedef LONG (FARPASCAL * DLLMAPSERVICES) (HGLOBAL ,LPVIPERVAR ,UINT ,PDWORD_PTR,UINT ,LPVIPERVAR ,UINT ,LPVIPERVAR ,UINT ,LONG);

typedef struct
{
  LPBDS       lpbdsptr;                                   // pointer to lpBDS (internal use)
  LPRESULTVAR lprvptr;                                    // WIL resultvar (internal use)
  DWORD       dwLevel;                                    // feature interface level
  DWORD       dwSize;                                     // size of the structure
  DLLSERVICES lpfnDllServices;                            // address of DLLServices function
  HGLOBAL     hBatData;                                   // buffer pointed to by lpBDS
  // feature level 1002+
  HANDLE      hImpersonatedUser;                          // handle to impersonated user token
  // feature level 1003+
  HWND        hParentWnd;                                 // window handle of parent application
  // feature level 1004+
  DLLVARHANDLER    lpfnDllVarHandler;                     // address of DllVarHandler function
  DLLARRHANDLER    lpfnDllArrHandler;                     // address of DllArrHandler function
  DLLARRAYSERVICES lpfnDllArrayServices;                  // address of DllArrayServices function
  // feature level 1005+
  DLLSTOREERRORINFO lpfnDllStoreErrorInfo;                // address of DllStoreErrorInfo function
  // feature level 1006+
  DLLBINARYSERVICES lpfnDllBinaryServices;                // address of DllBinaryServices function
  // feature level 1007+
  // DllVarHandler now supports variable "pointers"
  // feature level 1008+
  // variable table not cleared until after extenders are shut down
  // feature level 1009+
  DLLVARHANDLERW           lpfnDllVarHandlerW;            // address of DllVarHandlerW function
  DLLARRAYSTRINGSTOANSI    lpfnDllArrayStringsToAnsi;     // address of DllArrayStringsToAnsi function
  DLLARRAYSTRINGSTOUNICODE lpfnDllArrayStringsToUnicode;  // address of DllArrayStringsToUnicode function
  // feature level 1010+
  DLLARRAYSWAPELEMENTS     lpfnDllArraySwapElements;      // address of DllArraySwapElements function
  // Feature level 1011+ (WB 2021b or later)
  DLLMAPSERVICES           lpfnDllMapServices;            // address of DllMapServers function
} VIPERSTRUCT, FAR *LPVIPERSTRUCT;

typedef struct
{
  DWORD dwDim[5];                      // array dimension sizes
  BYTE  cDims;                         // number of array dimensions
  BYTE  cFlags;                        // array flags
  DWORD dwElements;                    // total number of elements
  DWORD dwDataSize;                    // total size of array data
} DLLARRAYSERVICES1, FAR *LPDLLARRAYSERVICES1;

typedef int FAR PASCAL ViperCallback (int, LONG, LPVIPERSTRUCT, LPVIPERVAR);
typedef ViperCallback FAR *LPVIPERCALLBACK;

DllExport LONG FAR PASCAL WILExtender(HWND,
                                      HINSTANCE,
                                      LONG,
                                      short,
                                      lpAddOnCallback,
                                      LPSTR,
                                      LPLONG,
                                      LPSTR,
                                      LPSTR,
                                      LPSTR,
                                      LPSTR,
                                      LPSTR);

#ifdef UNICODE
  DllExport LONG FAR PASCAL WILExtender2W(HWND,
                                          HINSTANCE,
                                          LONG,
                                          short,
                                          LPVIPERCALLBACK,
                                          LPVIPERSTRUCT,
                                          LPLONG,
                                          LPVIPERVAR,
                                          LONG,
                                          LPVIPERVAR);
#else
  DllExport LONG FAR PASCAL WILExtender2(HWND,
                                         HINSTANCE,
                                         LONG,
                                         short,
                                         LPVIPERCALLBACK,
                                         LPVIPERSTRUCT,
                                         LPLONG,
                                         LPVIPERVAR,
                                         LONG,
                                         LPVIPERVAR);
#endif

#ifdef UNICODE
  DllExport LONG FAR PASCAL WILExtenderQueryW(short,
                                              LPVOID,
                                              LONG);
#else
  DllExport LONG FAR PASCAL WILExtenderQuery(short,
                                             LPVOID,
                                             LONG);
#endif

#define COMMEXT_WILSON  0x30030105L


//-------------------------------------------------------------------
// MISC. DEFINITIONS
//-------------------------------------------------------------------
#define FUNCTION        2
#define CONSTANT        3
#define RESULT_LONG     0
#define RESULT_STRING   1
#define RESULT_FLOAT    2

#define RESULT_CALLBACK_ERROR -4
#define RESULT_FATAL_ERROR    -3
#define RESULT_MODERATE_ERROR -2
#define RESULT_MINOR_ERROR    -1
#define RESULT_SUCCESS         0

//-------------------------------------------------------------------
//==== DELIMSIZE IS DEFINED BY INTERPRETER, DON'T CHANGE THIS! ======
//-------------------------------------------------------------------
#define DELIMSIZE  30
#define KTDELIMSIZE  30                // Gives us 30 character function names. NO modify


//-------------------------------------------------------------------
// COMMAND EXTENSION STRUCTURES
//-------------------------------------------------------------------
typedef struct _COMMEXTSTRUCT {
   short    ident;                     // Our Identifier for item...
   short    type;                      // (2=function,3=constant)
   LONG     param;                     // (funct ? num-args)
                                       // (const ? value)
   short    size;                      // Length of keytoken
   char     delim[DELIMSIZE + 1];      // key token
} COMMEXTSTRUCT, FAR *LPCOMMEXTSTRUCT;


#ifdef UNICODE
  typedef struct tableinfoW
  {
    LONG  ident;                         // delimindex - must be consecutive!!!!!!
    LONG  type;                          // 2=FUNCTION 3=CONSTANT 4=COMMAND
    INT64 ulMask;                        // 64-bit mask (16 params, 4 bits each)
    LONG  param1;                        // (funct ? hi(reserved) lo(hibyte(num-optional-args) lobyte(num-required-args)))   (constant ? value)
    short size;                          // length of function name
    WCHAR delim[KTDELIMSIZE + 1];        // funct/const name - must be lowercase!
  } tablestructW, FAR *LPTABLESTRUCTW;
#else
  typedef struct tableinfo
  {
    LONG  ident;                         // delimindex - must be consecutive!!!!!!
    LONG  type;                          // 2=FUNCTION 3=CONSTANT 4=COMMAND
    INT64 ulMask;                        // 64-bit mask (16 params, 4 bits each)
    LONG  param1;                        // (funct ? hi(reserved) lo(hibyte(num-optional-args) lobyte(num-required-args)))   (constant ? value)
    short size;                          // length of function name
    char  delim[KTDELIMSIZE + 1];        // funct/const name - must be lowercase!
  } tablestruct, FAR *LPTABLESTRUCT;
#endif


//-------------------------------------------------------------------
// Call Back Definitions
//
// Parameter - 1 (int)     -3 = Fatal error occured, error string in LPSTR
//                         -2 = Moderate error occured, error string in LPSTR
//                         -1 = Minor error occured, error string in LPSTR
//                          0 = Function suceeded, check BOOL flag to
//                              see if result is in LPSTR or LONG.
//                              (RESULT_LONG or RESULT_STRING).
//
// Return Values:    Same as first parameter (above), unless the
//                   callback routine has a problem, like not being
//                   able allocate storage space... in this case it
//                   will return -4.
//
// NOTE: Return values are to be specified by the 'RESULT_???_ERROR'
//       definitions under MISC. DEFINITIONS (above).
//-------------------------------------------------------------------
int FAR PASCAL COMMEXT_CALLBACK_RTN (int, LONG, LPSTR, BOOL, LONG, LPSTR);

typedef int FAR PASCAL COMMEXT_CALLBACK (int, LONG, LPSTR, BOOL, LONG, LPSTR);
typedef COMMEXT_CALLBACK FAR *LPCOMMEXT_CALLBACK;


//#define MBSS(p1, p2) {MessageBox(GetFocus(), (p2), (p1), MB_OK | MB_SETFOREGROUND);}
//#define MBSN(p1, p2) {char szDebug[20]; MyLtoa((p2), szDebug, 10); MessageBox(GetFocus(), szDebug, (p1), MB_OK | MB_SETFOREGROUND);}

#endif
