#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//////////////////////////////////////////////////////////////////////////////////
// Set std:string type.
#ifdef UNICODE
typedef std::wstring tstring;
#else
typedef std::string tstring;
#endif

/////////////////////////////////////////////////////////////////////////////////
// Map iterator cache class
class CIterCache {

public:
	CIterCache() : m_nFirstIndex(21221), m_nMaxIndices(m_nFirstIndex+101) {} 
	~CIterCache(){}

	int AddIterator(DWORD_PTR dwMapIterator) throw(...)
	{
		int nIndex = m_nFirstIndex;

		for (; nIndex < m_nMaxIndices; nIndex++) {
			if ( m_IterTable.count(nIndex) <= 0 ) { 
				m_IterTable[nIndex] = dwMapIterator;
				break;
			}
		}
		
		return nIndex;
	}

	DWORD_PTR GetIterator( int nIndex) throw(...) 
	{
		DWORD_PTR dwReturn = 0;
		
		if ( m_IterTable.count(nIndex) > 0 )  
			dwReturn = m_IterTable[nIndex];
		return dwReturn;
	}

	void UpdateIterator( int nIndex, DWORD_PTR dwMapIterator ) throw(...)
	{
		m_IterTable[nIndex] = dwMapIterator;
	}

	void RemoveIterator(int nIndex) throw(...)
	{
		std::map<int,DWORD_PTR>::iterator it;

		it = m_IterTable.find(nIndex);
		if ( it != m_IterTable.end() )
			m_IterTable.erase(it);
	}

	BOOL IsIndexInBounds(int nIndex) noexcept 
	{
		return nIndex >= m_nFirstIndex && nIndex <= m_nMaxIndices;
	}
protected:
	const int m_nFirstIndex;
	const int m_nMaxIndices;
	std::map<int, DWORD_PTR> m_IterTable; // Table of open map iterators.
};

/////////////////////////////////////////////////////////////////////////////////
// The 64-bit integer math extender class.
class CMapEx : public CWilExtenderT<CMapEx> 
{
public:
	///////////////////////////////////////////////////
	// ctor/dtor
	CMapEx() : dwLastError(0){}
	~CMapEx(){}

	///////////////////////////////////////////////////
	// Required Extender class members.
	LONG Identity( LPVOID lpvInfo, LONG   nCount);
	LONG InterfaceVersion(LPVOID lpvInfo, LONG nCount);
	LONG Initialization(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);
   LONG LoadTable(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);
   LONG ShutDown(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);

	////////////////////////////////////////////////////
	// Extender WIL function implementations.
	LONG mpTest(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount); 
	LONG mpSetHash(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);
	LONG mpArrayize(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);
	LONG mpMapize(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);

protected:

	///////////////////////////////////////////////////
	// Utility members.
	BOOL IsMap(LPVIPERVAR lpvvArg) noexcept;
	inline BOOL IsString(LPVIPERVAR lpvvArg) noexcept
	{
		return lpvvArg && (lpvvArg->inittype & (VARTYPE_STRING|VARTYPE_UNICODE) && lpvvArg->lpwstr);
	}
	inline LPTSTR GetTstr(LPVIPERVAR pvVar) noexcept
	{
#ifdef UNICODE
		if ( pvVar && pvVar->inittype & VARTYPE_UNICODE )
			return pvVar->lpwstr;
#else
		if ( pvVar && pvVar->inittype & VARTYPE_STRING )
			return pvVar->lpstr;
#endif
		return NULL;
	}

	inline BOOL IsInterfaceOk() noexcept
	{
		return m_lpViperStruct->dwLevel >=  EXTENDER_FEATURE_LEVEL;
	}

	BOOL StrToInt( LPTSTR pszNumber, int nBase, int * pnOut ) noexcept;

	///////////////////////////////////////////////////
	// Properties.
	DWORD dwLastError;             // Extender error number. 
	CIterCache m_Cache;            // Iterator index cache.
};



/////////////////////////////////////////////////////////////////////////////////
// Set RTlibrary to user's locale.
class CUserLocale {
public:
	CUserLocale()
	{
		SetToEnvironment();
	}

	~CUserLocale()
	{
		Restore();
	}
	void SetToEnvironment()
	{
		// Save current locale.
		LPTSTR psztTmp = _tsetlocale( LC_ALL , NULL);
		if (psztTmp) Default = psztTmp;

		// Set the locale to environment's default locale.
		_tsetlocale( LC_ALL , _T("")); 
	}
	void Restore()
	{
		if ( !Default.empty() ) {
			_tsetlocale( LC_ALL, Default.c_str());
			Default.clear();
		}
	}

protected:
	tstring Default;
};

/////////////////////////////////////////////////////////////////////////////////
// Wrapper class for DllMapServices function error return exceptions. 
class CDllMapException {
private:
   CDllMapException(): m_nDllErr(0) {}  // Avoid uninitialized construction.
   int m_nDllErr; // Dll Error.
public:
   CDllMapException( CDllMapException& e) : m_nDllErr(e.m_nDllErr) {}  // Copy constructor.
   CDllMapException( int n) : m_nDllErr(n) {}                          // Initializing constructor.
   ~CDllMapException() {}
   int DllErr() { return m_nDllErr; }
	int ResourceId() {
		if ( m_nDllErr ) 
			return (IDS_DLLERROR1-1) + m_nDllErr;
		else
			return 0;
	}
};

/////////////////////////////////////////////////////////////////////////////////
// Wrapper class for generic DLL function error return exceptions.
class CDllGenericExecption {
private:
   CDllGenericExecption(): m_nDllErr(0){}  // Avoid uninitialized construction.
   int m_nDllErr; // Dll Error.
	string m_strText; // Error message usually DLL function name.
public:
	CDllGenericExecption ( CDllGenericExecption & e) : m_nDllErr(e.m_nDllErr), m_strText(e.m_strText) {}  // Copy constructor.
   CDllGenericExecption ( int n, string &strText) : m_nDllErr(n), m_strText(strText) {}                  // Initializing constructor.
   ~CDllGenericExecption () {}
   int DllErr() { return m_nDllErr; }
	const char * what() {
		CHAR szTmp[32];
		if ( !_itoa_s(m_nDllErr, szTmp, _countof(szTmp), 10) ) { // Note: no locale changes.
			m_strText += ' ';
			m_strText += szTmp;
		}
		return  m_strText.data();
	}
};

/////////////////////////////////////////////////////////////////////////////////
// Wrapper class for CRT's errno. Use to throw errno exceptions.
class CErrnoException {
private:
   CErrnoException() : m_nErrno(0) {}  // Avoid uninitialized construction.
   int m_nErrno; // CRT errno value.
public:
   CErrnoException( CErrnoException& e) : m_nErrno(e.m_nErrno) {}  // Copy constructor.
   CErrnoException( int n) : m_nErrno(n) {}                        // Initializing constructor.
   ~CErrnoException() {}
   int GetErrno() { return m_nErrno; }
};

/////////////////////////////////////////////////////////////////////////////////
// Wrapper class for user error exceptions. 
class CUserException {
private:
   CUserException() : m_nUserErr(0) {}  // Avoid uninitialized construction.
   int m_nUserErr;      // User Error resource ID.
public:
   CUserException( CUserException& e) : m_nUserErr(e.m_nUserErr) {}  // Copy constructor.
   CUserException( int n) :m_nUserErr(n) {}                          // Initializing constructor.
   ~CUserException() {}
   int UserErr() { return m_nUserErr; }
	int ResourceId() {
		if ( m_nUserErr ) 
			return m_nUserErr;
		else
			return 0;
	}
};

