// WilExtender.h: interface for the CWilExtender class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_WILEXTENDER_H__6EB69207_FCE5_4830_AC87_F807CE17AAA2__INCLUDED_)
#define AFX_WILEXTENDER_H__6EB69207_FCE5_4830_AC87_F807CE17AAA2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


///////////////////////////////////////////////////////////////////////
//  Handy datatype definitions.
///////////////////////////////////////////////////////////////////////
#ifndef MAX_PATH 
#define MAX_PATH 256
#endif
#define INTERFACELEVEL  4              // Extender interface level (DON'T CHANGE!)
#define EXTENDER_FEATURE_LEVEL 1004    // Required BATDLL feature interface level
#define EXTENDERVERSION 10002          // DLL version number (just a default)
#define MAX_EXT_NAME MAX_PATH          // Max extender name size.
#define MIN_MESSAGE_BUFFER 0x200       // Minimum size of string resource buffer.

// WIL extender includes - do NOT remove.
#pragma pack(push,4)
extern "C" {
#include "wwwpform.h"
#include "addons.h"
}
#pragma pack(pop)

///////////////////////////////////////////////////////////////////////
// Each extender function should have a corresponding class method.
// The class method must have the following 5 parameters:
//
// HWND            hWnd                    // hWnd of parent application
// HINSTANCE       hInst                   // hInstance of parent application
// LPLONG          lpLong                  // pointer to your 32 bits
// LPVIPERVAR      lpvvArg                 // pointer to parameter array
// LONG            nArgCount               // parameter count for lpvvArg
// 
// and it must return BOOL. The typdef below defines a pointer
// to a function with these parameters and return type.
//
////////////////////////////////////////////////////////////////////////
class CWilExtender;  // Forward declare.
typedef BOOL (CWilExtender::*PMEMBER) (HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount);  

///////////////////////////////////////////////////////////////////////
// Command map stucture.
///////////////////////////////////////////////////////////////////////
typedef struct tagCommandMap{  
   tablestruct t;             // Extender Command table
   PMEMBER pMember;           // Pointer to supporting member.
} COMMAND_MAP, * LPCOMMAND_MAP;

///////////////////////////////////////////////////////////////////////
//  End datatype definitions.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
// Command table map macros.
//
// Use these macro's to create the command table and function map
// so WIL knows about your extenders functions. 
// 
///////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: BEGIN_COMMAND_MAP
 *
 * Purpose: Begins initialization of extender's command table. 
 *
 * Arguments: _THISCLASS  - Name of your class.
 *            _THISOBJECT - Your extender class object.
 *
 * Example: CMyExtender MyExtender;
 *          BEGIN_COMMAND_MAP(CMyExtender, MyExtender)
 *
 ********************************************************************/
#define BEGIN_COMMAND_MAP(_THISCLASS, _THISOBJECT) \
   CWilExtender * CWilExtender::m_pThisObject = dynamic_cast<CWilExtender *>(&_THISOBJECT); \
   const LPCOMMAND_MAP  _THISCLASS::GetCommandMap() const \
		{ return (LPCOMMAND_MAP) _THISCLASS::CommandMap; } \
   const COMMAND_MAP _THISCLASS::CommandMap[] = \
	{ \

/********************************************************************
 * Name: COMMAND_MAP_FUNCTION
 *
 * Purpose: Adds extender function to command table. 
 *
 * Arguments: _PARM_TYPES - function's paremeter types (up to 16)
 *            _NUM_PARAMS - number of parameters function accepts.
 *            _NAME       - name used by scripts to call function.
 *            _MEMBER     - member of your class.
 *                           Any class member used here must be 
 *                           like the following:
 *                           BOOL CMyExtender::MyFunc( HWND hWnd, 
 *                                                     HINSTANCE hInst, 
 *                                                     LPLONG lpLong, 
 *                                                     LPVIPERVAR lpvvArg, 
 *                                                     LONG nArgCount)
 *                           where "CMyExtender" is the name of your
 *                           class and "MyFunc" is any name you choose
 *                           for your class member function. (static 
 *                           funtions can not be used here.)
 *
 * Example: COMMAND_MAP_FUNCTION(FIVE_STRING_PARMS,5,"myfunc",CMyExtender::MyFunc)
 *
 ********************************************************************/
#define COMMAND_MAP_FUNCTION( _PARM_TYPES, _NUM_PARAMS, _NAME, _MEMBER) \
{0, FUNCTION, 0x0000000000000000|(_PARM_TYPES), 0x00000000L|(_NUM_PARAMS), 0, _NAME,(PMEMBER) (&_MEMBER) }, \

/********************************************************************
 * Name: COMMAND_MAP_CONSTANT
 *
 * Purpose: Adds extender constant to command table. 
 *
 * Arguments: _VALUE - constants value.
 *            _NAME  - constants name.
 *
 * Example: COMMAND_MAP_CONSTANT(1, "@SUCCESS")
 *        
 ********************************************************************/
#define COMMAND_MAP_CONSTANT( _VALUE, _NAME) \
{0, CONSTANT, 0x0000000000000000, 0x00000000L|(_VALUE), 0, _NAME,(PMEMBER) (NULL) }, \

/********************************************************************
 * Name: END_COMMAND_MAP()
 *
 * Purpose: End command table initialization.
 *
 * Arguments: None.
 *
 * Example: CMyExtender MyExtender;
 *          BEGIN_COMMAND_MAP(CMyExtender, MyExtender)
 *          COMMAND_MAP_FUNCTION(FIVE_STRING_PARMS,5,"myfunc",CMyExtender::MyFunc)
 *          END_COMMAND_MAP()
 *
 ********************************************************************/
#define END_COMMAND_MAP() \
   {-1, 0, 0x0000000000000000, 0x00000001L, 0, "", (PMEMBER) NULL } \
   }; \

/********************************************************************
 * Name: DECLARE_COMMAND_MAP()
 *
 * Purpose: Place this in your class definition to declare the
 *          class's command table and accessing method.
 *
 * Example: class CMyExtender : public CWilExtender  
 *          {
 *              public:
 *              CMyExtender();
 *	             virtual ~CMyExtender();
 *                       
 *              //  Declare command table support with this macro. 
 *              DECLARE_COMMAND_MAP() 
 *            }
 *
 ********************************************************************/
#define DECLARE_COMMAND_MAP() \
public: \
   static const  COMMAND_MAP   CommandMap[]; \
   virtual const LPCOMMAND_MAP GetCommandMap() const; \

// _PARM_TYPES value for 5 string parameters.
#define FIVE_STRING_PARMS 0x0000000000011111 

/********************************************************************
 * Name: PARAM_TYPE 
 *
 * Purpose: Handy for setting COMMAND_MAP_FUNCTION macro's _PARM_TYPES argument. 
 *          Use one per function parameter all "OR"ed (|) together.
 *
 * Arguments: _PARAM  = the parameter type to set. Must be between 1 and 16.
 *            _TYPE   = datatype of _PARAM (see defines below.) 
 *
 * Example: COMMAND_MAP_FUNCTION( PARAM_TYPE(1,INTTYPE)|PARAM_TYPE(2,STRTYPE),2,...
 *
 ********************************************************************/
/* bug fix on PARAM_TYPE macro Oct 2005 */
#define PARAM_TYPE( _PARAM, _TYPE) (((0xffffffffffffffff&((_TYPE)&0xf)))<<(((_PARAM)-1)*4))
// The short version of above
#define PT(_PARAMX, _TYPEX) PARAM_TYPE(_PARAMX, _TYPEX)

// Datatypes of parametesr a WIL extender function can have.
// (values for _TYPE above.)
#define INTTYPE   0x0   // = int
#define STRTYPE   0x1   // = string
#define FLOATTYPE 0x2   // = float
#define ARRAYTYPE 0x3   // = array
#define VTTYPE    0x4   // = variant
#define ASISTYPE  0xf   // = WIL makes its best guess.


///////////////////////////////////////////////////////////////////////
// End command map macros.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
// The WIL extender class. (At last...)
///////////////////////////////////////////////////////////////////////
class CWilExtender  
{
public:
   CWilExtender();
	virtual ~CWilExtender();
     
public:
   
   ///////////////////
   // Member functions.
   VOID * GetWorkBuffer(DWORD dwSize = 0x100, BOOL bSaveContent = FALSE);
   DWORD GetWorkSize() {return m_dwWorkSize;};
   LPTSTR MyLastSlashInString(LPTSTR lpString);

   // Gives access to derived class's command table.
   virtual const LPCOMMAND_MAP GetCommandMap() const = 0; // Note: pure virtual.
   
   // Required members for extender startup and shutdown.
   virtual BOOL Initalization(HWND hWnd, 
                              HINSTANCE hInst, 
                              LPLONG lpLong, 
                              LPVIPERVAR lpvvArg, 
                              LONG nArgCount);
   virtual BOOL LoadTable(HWND hWnd, 
                          HINSTANCE hInst, 
                          LPLONG lpLong, 
                          LPVIPERVAR lpvvArg, 
                          LONG nArgCount);
   virtual BOOL ShutDown(HWND hWnd, 
                         HINSTANCE hInst, 
                         LPLONG lpLong, 
                         LPVIPERVAR lpvvArg, 
                         LONG nArgCount);
   
   // Use these to return values to the WIL Interpreter.
   virtual BOOL ReturnResult( LONG lNumber );
   virtual BOOL ReturnResult( int nNumber ) { return ReturnResult( (LONG) nNumber );};
   virtual BOOL ReturnResult( double dNumber );
   virtual BOOL ReturnResult( LPTSTR lpStr );
   virtual BOOL ReturnResult(LPVARIANT lpVariant);
   virtual BOOL ReturnResult( LPVIPERVAR lpVv);
   virtual BOOL ReturnError( int nSeverity, int nErrorID);
   virtual BOOL ReturnCancel(){return -5;};

   // Cover members for variable and array handlers.
   virtual LONG DllVarHandler(LPSTR lpVarName, LPVIPERVAR lpVipVar, LONG nRequest)
   {
      return m_lpViperStruct->lpfnDllVarHandler(m_lpViperStruct->hBatData,
                                                lpVarName,
                                                lpVipVar,
                                                sizeof(vipervar),
                                                nRequest);
   }

   virtual long DllArrHandler(LPVIPERVAR lpArr, DWORD dwSub1, DWORD dwSub2, DWORD dwSub3, DWORD dwSub4, DWORD dwSub5, LPVIPERVAR lpVipVar, LONG nRequest)
   {
      return m_lpViperStruct->lpfnDllArrHandler(m_lpViperStruct->hBatData,
                                                lpArr,
                                                dwSub1,
                                                dwSub2,
                                                dwSub3,
                                                dwSub4,
                                                dwSub5,
                                                lpVipVar,
                                                sizeof(vipervar),
                                                nRequest);
   }

   virtual long DllArrayServices(LPVIPERVAR lpArr, LPBYTE lpBuf, UINT uBufSize, LONG nRequest)
   {
      return m_lpViperStruct->lpfnDllArrayServices(m_lpViperStruct->hBatData,
                                                   lpArr, 
                                                   sizeof(vipervar), 
                                                   lpBuf, 
                                                   uBufSize, 
                                                   nRequest);
   }

   // Utility method used internally.
   virtual int LoadMyExtString(int wID, LPTSTR lpBuffer, int nBuffermax);

   /////////////////////
   // Member properties.
   static CWilExtender * m_pThisObject;      // Gives global functions access to derieved object.
   static HINSTANCE      m_hInstDLL;         // DLL instance handle.

   LONG            m_ViperInterfaceVersion;  // WIL interface version.
   LPVIPERCALLBACK m_lpViperCallback;        // Pointer to return value callback function.
   LPVIPERSTRUCT   m_lpViperStruct;          // Pointer to information structure.
   LPVIPERVAR      m_lpViperReturn;          // Pointer to return value structure.
   short           m_CmdId;                  // Identifies extender function being called.
   TCHAR           m_szExtName[MAX_EXT_NAME];// Name of extender.
   DWORD           m_dwExtVersion;           // Extender dll version.

private:
   VOID *          m_lpWork;                 // Work buffer.
   DWORD           m_dwWorkSize;             // Size of work buffer.
   
};

#endif // !defined(AFX_WILEXTENDER_H__6EB69207_FCE5_4830_AC87_F807CE17AAA2__INCLUDED_)
