/*===========================================================================*/

#define EXTENDERVERSION 41001          // DLL version number

#define INTERFACELEVEL  4                // extender interface level (DON'T CHANGE!)
#define EXTENDER_FEATURE_LEVEL 1004      // extender feature level (DON'T CHANGE)

#define STRICT                         // only if you prototype carefully

/*===========================================================================

                                     WILX.C
                             WILX WIL Extender DLL

                  Copyright 1994-2004 Wilson WindowWare, Inc.

  ---------------------------------------------------------------------------
                                Revision History
  ---------------------------------------------------------------------------

  Rev 10000   Apr 08, 1994   MHW
    Initial release.

  Rev 11000   Oct 09, 1994   REM
    Reformatted and cleaned up code and comments.
    Removed 'xstrcpy' and associated references
      (apparently MHW overlooked the 'lstrcpyn' WINAPI function).
    Removed silly demo functions, and added 10 useful functions:
      xExtenderInfo
      xBaseConvert
      xCursorSet
      xDiskLabelGet
      xDriveReady
      xGetElapsed
      xHex
      xMemCompact
      xMessageBox
      xVerifyCCard

  Rev 11100   Oct 12, 1994   REM
    Added functions:
      xGetChildHwnd
      xSendMessage

  Rev 11101   Jan 05, 1995   REM
    Converted from Borland C++ to Microsoft Visual C++.

  Rev 11102   Apr 10, 1995   REM
    Changed xDriveReady to support all drive types, not just floppies.
    Fixed a problem with xDriveReady sometimes causing GP faults.
    xDriveReady no longer uses assembly code.

  Rev 11103   Apr 14, 1995   REM
    Removed constant @LF (it's been added to WIL).
    Added   constant @MBOKCANCEL.
    Added   constant @MBYESNO.
    Improved data type naming conventions.
    Removed FindNext loop in xDiskLabelGet (since Windows requires DOS 3+).
    Improved FindFirst function.

  Rev 11104   May 02, 1995   REM
    Fixed problem with xGetChildHwnd, if "text" was a null string.

  Rev 11105   Aug 24, 1996   REM
    Added support for localization of language strings.
    Changed xMemCompact to use HeapCompact in 32-bit version.

  Rev 11106   Jan 14, 1997   REM
    Added xEjectMedia.
    Fixed problem with xDriveReady always returning @TRUE in Windows NT.

  Rev 11107   Oct 22, 1998   REM
    xEjectMedia no longer returns an error under Windows NT if drive is empty.

  Rev 11108   Jan 05, 1999   REM
    xMessageBox will now come to the foreground in Windows 98 and NT 5.0.

  Rev 31000   Nov 01, 2000   REM
    Removed 16-bit code.  Extender is now 32-bit only.
    Removed xCursorSet (not useful in 32-bit Windows).
    Removed xMemCompact (not useful in 32-bit Windows).
    Converted to WilExtender version 2 interface.
    Fixed problem with xGetChildHwnd if 'child-text' was blank.

  Rev 41000   Mar 30, 2004   MW
    Finished conversions to 44I 2004B+ series extenders

  Rev 41001   Sep 07, 2011  
    64-bit port.

  ---------------------------------------------------------------------------
                                  Description
  ---------------------------------------------------------------------------

  This is an example of a WIL extender.  A WIL extender is an external DLL
  designed to complement the WIL script processor.  The WIL Processor has
  an "AddExtender" function that will load a WIL extender.  It basically
  performs a Windows API "LoadLibrary" on the extender, and then proceeds
  to extract its capabilities by a series of calls:

    First it calls the WILQueryExtender function case 2 to negotiate a
    parameter interface level and to decide it the WIL engine and the
    extender DLL are compatible.  Old-style extenders that do not support
    the  WILQueryExtender function are assumed to be interface level 1.
    This version of the sample supports interface level 2.

    Then, assuming the Extender supports interface level 2, a call to the
    WILExtender2 function, case -1 is performed to determine the number of
    functions, constants, commands and other table items that the extender
    wishes to add to the WIL command table.

    Next a series of calls to the same WILExtender2 function with case -2
    is performed to copy table entries to the main WIL engine tables.

  The WIL interpreter provides a pointer to a 32 bit memory location (LONG
  FAR *).  This pointer provides 32 bits of storage associated with the
  currently running script.  Different scripts will point to a different 32
  bit location.  This allows (if required) state information to be saved by
  the extender.  In addition, the extender may allocate a global memory
  block for storage, and save the handle of the memory block in the memory
  location.  This allows almost unlimited local data for the extender.

  When the WIL script processor shuts down, WILExtender2 case -3 will be
  called once to allow the extender to perform clean-up processing.

  WILExtenderQuery case 1 is designed for interrogation of the extender to
  determine what it is.

  The CommandExtTable defines the commands that the extender responds to,
  and the case #'s that correspond with the commands.

  Up to 16 parameters may be passed to the extender.  Each parameter is
  passed as a vipervar structure containing an integer (long), a string, and
  a floating point member, as well as an inittype member indicating what
  type of parameter it actually is (ie, which member(s) of the structure
  contain valid values):

    typedef struct vipervariable
    {
      double  flt;                     // 8-byte floating point number
      LONG    x;                       // if int variable, value is here
      int     inittype;                // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
      union
      {
        LPSTR lpstr;                   // pointer to string in string table
        DWORD handle;                  // handle for miscellaneous object types
      };
    } vipervar, FAR *LPVIPERVAR;

  inittype will be one (or more) of the following:

    VARTYPE_UNDEF       0              // undefined (ie, beyond nArgCount)
    VARTYPE_INT         1              // integer value (in 'x')
    VARTYPE_STRING      2              // string (pointed to by 'lpstr')
    VARTYPE_FILE        5              // FILE and INT bits set (in 'x')
    VARTYPE_OLEOBJECT  17              // OLE AND INT bits set (in 'x')
    VARTYPE_FLOATNUM   32              // floating point value (in 'flt')
    VARTYPE_BINARY     65              // BINARY and INT bits set (in 'x')
    VARTYPE_ARRAY     256              // an array (complex type using all structure elements)

  Note that a variable can be a string and/or integer and/or float.

  Your function also receives a pointer to one of these structures to use
  for returning a value to the WIL DLL.

  In order to avoid using C runtime acsii<->string conversion functions, a
  pair of conversion routines (MyLtoa and MyAtol) has been provided.

  You should provide one function to return the version number of your
  extender.  Your version number should start at 10000 and increment as
  the DLL is modified.

  Adopt a naming convention so that all functions in the extender begin
  with the same group of letters.  A function name must be 30 characters or
  less.

  ---------------------------------------------------------------------------

  A number of macros have been provided to assist in returning values to
  the WIL interpreter.  These are:

    RETLONG(num)
      Returns a long integer as the return value of the function.

    RETFLOAT(floatnum)
      Returns a floating point value as the return value of the function.

    RETSTRING(string)
      Returns a string as the return value of the function.

    RETERROR(level, rcfilestringid)
      Returns an error level (use -1, -2, or -3).  We recommend -1 and a
      string from the resource file for the error.  The string itself
      should start with a 3-digit error number, followed by a colon.

    RETCANCEL
      Used if you pop up a dialog box and the user hits the Cancel button.

    ANSWERSTRING(string)
      Allows the extender to return a string response, but does not
      return to the WIL interpreter directly as the other functions do.
      It allows a response to be constucted in a global memory buffer,
      the buffer contents returned, and then allows processing to
      continue, usually to free the global buffer.

  ---------------------------------------------------------------------------

  One final word of advice for those of you who are new to the joys of
  writing DLL's.  Without getting into a lot of technical detail about
  memory segments and heaps and stacks, let's just say that all programs
  have a data segment (DS, also known as the local heap) and a stack
  segment (SS).  For EXE's, the data segment and stack segment are one and
  the same (DS == SS).  But DLL's don't get their own stack segment; they
  share the stack of the EXE that is calling them, and therefore the data
  segment and stack segment are different (DS != SS).

  That said, you need to keep these important points in mind:

  1. You must ALWAYS use 32-bit pointers to access objects that are on the
     stack, such as arrays and structures.

  2. You have no control over the size of the stack, which you may be
     sharing with other DLL's as well as with your calling EXE.  It's a
     good idea to avoid putting large arrays and the like on the stack.

  3. C Storage Classes 101.  Automatic variables (variables defined within
     functions, also known as local variables) get created on the stack:

       char szBuffer[200];             // goes on the stack

     "Static" variables get created on the heap instead:

       static char szBuffer[200];      // goes on the local heap

  If your compiler has a compile option to indicate that DS != SS, use it.
  It could warn you of possible pointer conversion problems.

     For Microsoft C (small model library), this is:

       -ASw

     For Borland C++, it's:

       -ms!

  ===========================================================================*/

/*---------------------------------------------------------------------------
                                    Includes
  ---------------------------------------------------------------------------*/

// Standard Windows include
#include <windows.h>
#include <winioctl.h>                  // for xEjectMedia

// Special platform-dependent includes
#include "wwwpform.h"
#include "addons.h"

/*---------------------------------------------------------------------------
                                    Defines
  ---------------------------------------------------------------------------*/

#define MAXCARDLEN 40                  // maximum credit card length

#define VWIN32_DIOC_DOS_IOCTL 1

#define SPI_GETFOREGROUNDLOCKTIMEOUT 0x2000
#define SPI_SETFOREGROUNDLOCKTIMEOUT 0x2001

// helpful functions for displaying debugging message boxes
#define MBSS(p1, p2) {MessageBox(GetFocus(), (p2), (p1), MB_OK | MB_SETFOREGROUND);}
#define MBSN(p1, p2) {char szDebug[20]; MyLtoa((p2), szDebug, 10); MessageBox(GetFocus(), szDebug, (p1), MB_OK | MB_SETFOREGROUND);}

#define MyToUpper(p1) LOBYTE(LOWORD((DWORD_PTR) (AnsiUpper((LPSTR) (DWORD_PTR) (BYTE) (p1)))))

// RETLONG passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETLONG(numb)         return (lpViperCallback(((lpViperReturn->x=numb)      +(lpViperReturn->inittype=VARTYPE_INT))      ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETFLOAT passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETFLOAT(numb)        return (lpViperCallback(((lpViperReturn->flt=numb)    +(lpViperReturn->inittype=VARTYPE_FLOATNUM)) ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETSTRING passes its argument back as the return value of the function,
// and also returns to the WIL Interpreter.

#define RETSTRING(str)        return (lpViperCallback(((lpViperReturn->lpstr=str)   +(lpViperReturn->inittype=VARTYPE_STRING))   ?RESULT_SUCCESS:RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETERROR(level,rcfilestringid) passes back an error massage to the WIL
// interpreter.
//   level MUST be -1, -2 or -3.
//     Use -1 for most run of the mill errors.
//         -2 for serious errors
//         -3 for absolutely fatal "we better die" errors
//   rcstringid is the string ID from the RC file that contains the error
//   message.  The first three characters of the string should be the same
//   as the string ID, and the fourth character should be a colon.

//#define RETERROR(sev, errnum) return (lpViperCallback(((LoadMyExtString(hinstDLL, (errnum), lpViperReturn->lpstr = szWork, sizeof(szWork)))+(lpViperReturn->inittype = VARTYPE_STRING)) ?(sev)         :(sev)         , ViperInterfaceVersion, lpViperStruct, lpViperReturn))
#define RETERROR(sev, errnum) return (lpViperCallback(((LoadString(hinstDLL, (errnum), lpViperReturn->lpstr = szWork, sizeof(szWork)))+(lpViperReturn->inittype = VARTYPE_STRING)) ?(sev)         :(sev)         , ViperInterfaceVersion, lpViperStruct, lpViperReturn))

// RETCANCEL returns a Cancel flag

#define RETCANCEL return (-5)

// ANSWERSTRING passes its string argument back as the function return, but
// does not return to the WIL program.  NOTE: The value that is returned by
// ANSWERSTRING must be the value you use in your own 'return' command.

#define ANSWERSTRING(str) lpViperCallback(  ((lpViperReturn->lpstr=str)+(lpViperReturn->inittype=VARTYPE_STRING))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERLONG(numb)  lpViperCallback(  ((lpViperReturn->x=numb)+(lpViperReturn->inittype=VARTYPE_INT))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define ANSWERFLOAT(numb) lpViperCallback(  ((lpViperReturn->flt=numb)+(lpViperReturn->inittype=VARTYPE_FLOATNUM))?RESULT_SUCCESS:RESULT_SUCCESS,ViperInterfaceVersion,lpViperStruct,lpViperReturn)
#define RETCANCEL return (-5)



/*---------------------------------------------------------------------------
                             Structures / Typedefs
  ---------------------------------------------------------------------------*/

typedef struct _DIOC_REGISTERS         // IOCTL structure, for xEjectMedia
{
  DWORD reg_EBX;
  DWORD reg_EDX;
  DWORD reg_ECX;
  DWORD reg_EAX;
  DWORD reg_EDI;
  DWORD reg_ESI;
  DWORD reg_Flags;
} DIOC_REGISTERS, *PDIOC_REGISTERS;

/*---------------------------------------------------------------------------
                              Function Prototypes
  ---------------------------------------------------------------------------*/

LONG  NEARPASCAL MyAtol(LPSTR);
void  NEARPASCAL MyLtoa(LONG, LPSTR, UINT);
BOOL  NEARPASCAL DebugMessage(HWND, LPSTR, LPSTR, BOOL);
LONG  NEARPASCAL Base2Dec(LPSTR, UINT);
LONG  NEARPASCAL Pow(UINT, UINT);
int   NEARPASCAL LoadMyExtString(HMODULE, int, LPSTR, int);
LPSTR NEARPASCAL MyLastSlashInString(LPSTR);

/*---------------------------------------------------------------------------
                                Global Variables
  ---------------------------------------------------------------------------*/

/* Superstitious data field to bypass/ignore bugs in certain versions of
   Windows that cause the beginning of the data area to be overwritten.
   This field should be the first data area defined and linked. */
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE hinstDLL;                    // DLL Instance saved

BOOL bWin95 = FALSE;                   // Windows 95+
BOOL bWin98 = FALSE;                   // Windows 98
BOOL bWinNT = FALSE;                   // Windows NT
BOOL bWin40 = FALSE;                   // Windows 95+ or Windows NT 4.0+
BOOL bWin50 = FALSE;                   // Windows NT 5.0+

/*---------------------------------------------------------------------------
  Command Extension Table for this DLL

  NOTE:   ALL STRINGS MUST BE RETURNED IN LOWER CASE.  Note also that in our
          table uppercase characters are used, but than in the case -2
          routine, the CharLower function is used to convert them to
          lower case

  NOTE:   The sizes of the string names CAN NOT exceed 30 characters.  Note
          also that in our table the sizes are specified as zero, and that
          the case -2 routine uses lstrlen to compute the size for the return

  Definitions
    ID       Number, starting from 0, incrementing by 1, terminated by -1.
             For functions, it is also the 'case' number used to access the
             function.

    TYPE     Either FUNCTION or CONSTANT.

    MASK     Bitmask indicating the types of the 16 parameters.  Each
             nibble (4 bits) represents one parameter, with the low
             nibble representing param1, and then continuing on.  The
             possible types are:

               0  integer
               1  string
               2  floating point number
               3  array
               F  as-is (don't validate or convert)

    PARAMS/  For constants, this is the actual constant value: long integer
    VALUE    constants only.  For functions, the low byte of the low word
             is the number of required parameters, the high byte of the low
             word is the number of (additional) optional paremeters, and
             the high word is reserved (and should be set to 0).

    SIZE     Set to 0.

    NAME     Name of the function or constant.
             MUST START WITH AN ALPHABETIC CHARACTER.  NO SPACES.
             30 CHARACTERS OR LESS.  USE A NAMING CONVENTION.

  The extender is not called for constants.  Constant value is determined
  and saved during extender initialization.
  ---------------------------------------------------------------------------*/

tablestruct CommandExtTable[] =
{
/*---------------------------------------------------------------------------
  In this example, I've reserved ID's 0 - 9 for constants.  This allows a
  nicer numbering convention for the three-digit error message strings.

  Although it would be nice to define lots of constants for message box
  styles and cursor types and such, don't forget that your user can only
  have a maximum of 500 items, combined, from all the extenders that are
  loaded.  Constants do count towards this limit.  Try to be considerate.

  Also, note that only integer constants are supported.
  ---------------------------------------------------------------------------*/

//----------------------------------------------------------------------
// ID   Type       Mask               Params/Value Size Name
//----------------------------------------------------------------------
  { 0,  CONSTANT,                   0, MB_OKCANCEL,  0, "@MBOKCANCEL"  },  // is equal to 0x0001
  { 1,  CONSTANT,                   0, MB_YESNO,     0, "@MBYESNO"     },  // is equal to 0x0004
//
  {10,  FUNCTION,  0x0000000000000000, 0x00000001L,  0, "xExtenderInfo"},  // xExtenderInfo(i)
  {11,  FUNCTION,  0x0000000000000001, 0x00000003L,  0, "xBaseConvert" },  // xBaseConvert(s,i,i)
  {13,  FUNCTION,  0x0000000000000001, 0x00000001L,  0, "xDiskLabelGet"},  // xDiskLabelGet(s)
  {14,  FUNCTION,  0x0000000000000001, 0x00000001L,  0, "xDriveReady"  },  // xDriveReady(s)
  {15,  FUNCTION,  0x0000000000000011, 0x00000002L,  0, "xGetElapsed"  },  // xGetElapsed(s,s)
  {16,  FUNCTION,  0x0000000000000001, 0x00000001L,  0, "xHex"         },  // xHex(s)
  {18,  FUNCTION,  0x0000000000000011, 0x00000003L,  0, "xMessageBox"  },  // xMessageBox(s,s,i)
  {19,  FUNCTION,  0x0000000000000001, 0x00000001L,  0, "xVerifyCCard" },  // xVerifyCCard(s)
  {20,  FUNCTION,  0x0000000000000010, 0x00000003L,  0, "xGetChildHwnd"},  // xGetChildHwnd(i,s,i)
  {21,  FUNCTION,  0x0000000000000000, 0x00000004L,  0, "xSendMessage" },  // xSendMessage(i,i,i,i)
  {22,  FUNCTION,  0x0000000000000001, 0x00000001L,  0, "xEjectMedia"  },  // xEjectMedia(s)
//--------------------------------------------------------------------------
  {-1,         0,                   0,  0x00000000,  0, ""             }
};

/*===========================================================================
                                    DllMain
  ===========================================================================*/

BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{
  OSVERSIONINFO osvi;

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);

  osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&osvi);

  if (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
    bWin95 = TRUE;
  else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT)
    bWinNT = TRUE;

  if ((bWin95) && (osvi.dwMajorVersion == 4) && (osvi.dwMinorVersion >= 10))
    bWin98 = TRUE;

  if (osvi.dwMajorVersion >= 4)
    bWin40 = TRUE;

  if (osvi.dwMajorVersion >= 5)
    bWin50 = TRUE;

  return (TRUE);
}

/*===========================================================================
                               WILExtenderQuery

  WILExtenderQuery is an required entry point.  It is used
  to provide helpful information to the WIL DLL and possibly other callers.
  ===========================================================================*/

DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  static char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*---------------------------------------------------------------------------
  case 1 is used to identify yourself to a caller.

  If lpvInfo is not NULL, it points to a character buffer where you may copy
  a string describing the extender.

  nCount is the size of the buffer pointed to by lpvInfo, in bytes.  When
  called by the WIL DLL, you can expect this to be a 256-byte buffer.

  The return value should be the version number of the extender.
  ---------------------------------------------------------------------------*/
    case 1:                            // version information
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadMyExtString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }

      return (EXTENDERVERSION);
    }

/*---------------------------------------------------------------------------
  case 2 is used to inform the WIL DLL of your preferred interface level.

  lpvInfo points to an array of LONG's, indicating the interface levels
  supported by the WIL DLL (each element represents one level).

  nCount indicates the number of elements in the array pointed to by lpvInfo.

  The return value must be the same as one of the elements in the lpvInfo
  array.  Currently, the WIL DLL supports interface levels 1 and 2, so your
  return value should be 1 or 2.  Newly designed extenders (such as WILX)
  use interface level 2 and have a WilExtender2 entry point.  However, it
  is possible to design an extender to support multiple interface levels.

  If you return an invalid value, the extender will not be loaded and the
  AddExtender command will fail.
  ---------------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }

/*---------------------------------------------------------------------------
  Return 0 for any unknown requests, or any requests you don't handle!
  ---------------------------------------------------------------------------*/
    default:
      return (0);
  }
}

/*===========================================================================
                                  WILExtender2

  This is a required entry point.  It is called each time a function in your
  extender is executed (identified by CmdId's >= 0).  In addition, it is
  called to perform various initialization and un-initialization tasks
  (identified by CmdId's < 0).
  ===========================================================================*/

DllExport LONG FAR PASCAL WILExtender2
  (HWND            hWnd,                  // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{
  LPSTR  lpMyIni = "www-prod.ini";        // our INI file
  LPSTR  lpMySec = "WILX";                // our INI file section
  static char szWork[256];                // some local storage
  UINT   uDebugLevel;
  BOOL   bDebugMonitor;


  if (CmdId >= 0)                         // don't do this for initialization calls
  {
    // controls verbosity of debugging output: 0 = none
    uDebugLevel = GetPrivateProfileInt(lpMySec, "DebugLevel", 0, lpMyIni);

    // controls whether debug messages go to a secondary debug monitor
    bDebugMonitor = GetPrivateProfileInt(lpMySec, "DebugMonitor", 0, lpMyIni);
  }

  switch (CmdId)                       // the "big switch" (cases < 0 are at the bottom)
  {

/*---------------------------------------------------------------------------
  Case -1 is called once, when your extender is loaded.  It is the place to
  perform your one-time initialization tasks.  The return value should be the
  number of items in your CommandExtTable (functions and constants).  You may
  also return 0 if you are unable to determine the correct number of items.

  You may return a negative value (-1) to indicate that your extender was
  unable to successfully initialize, for example, if it is running on an
  unsupported platform or a dependent DLL could not be loaded.  This will
  cause the AddExtender command to fail.  You may additionally provide a
  string indicating the reason for the failure:

    lpvvArg points to a character buffer into which you may copy an error
    message.

    nArgCount is the size of the buffer pointed to by lpvvArg, in bytes.
    You can expect this to be a 256-byte buffer.
  ---------------------------------------------------------------------------*/
    case -1:                           // Initialization
    {
      BOOL bInterfaceOk;
      bInterfaceOk=FALSE;
      if (lpViperStruct->dwLevel >= EXTENDER_FEATURE_LEVEL)
      {
        bInterfaceOk = TRUE;
        //lpfnDllServices = lpViperStruct->lpfnDllServices;
        //hBatData = lpViperStruct->hBatData;
      }


      if (bInterfaceOk==TRUE)
       {
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1);  // 1 extra at end
       }
      else
         {
		 lstrcpy((LPSTR)lpvvArg,"Need newer version of WIL DLL");
         return (-1);
         }
    }

/*---------------------------------------------------------------------------
  Case -2 is called repeatedly by the WIL extender.  On each call, one
  CommandTable entry is passed back to the WIL extender.  It is unlikely that
  you will need to modify this code.
  ---------------------------------------------------------------------------*/
    case -2:                           // Loading table
    {
      LPTABLESTRUCT lpTable;
      UINT          j;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      j = (UINT) nArgCount;

      if (CommandExtTable[j].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[j].ident;
        lpTable->type   = CommandExtTable[j].type;
        lpTable->param1 = CommandExtTable[j].param1;
        lpTable->ulMask = CommandExtTable[j].ulMask;
        lpTable->size   = lstrlen((LPSTR) CommandExtTable[j].delim);
        CharLower(CommandExtTable[j].delim);
        lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[j].delim);
      }

      return (1);
    }

/*---------------------------------------------------------------------------
  Case -3 is called during the WIL shutdown process.  If you need to do any
  cleanup processing, do it here.
  ---------------------------------------------------------------------------*/
    case -3:                           // shutdown
    {
      return (1);
    }





/*---------------------------------------------------------------------------
                            xExtenderInfo(i:request)

  Returns information on this extender
    request: 0 = version, 1 = total functions, 2 = total constants
  ---------------------------------------------------------------------------*/
    case 10:
    {
      LONG nRequest;
      UINT i, uCount, uType;

      nRequest = lpvvArg[0].x;

      if (nRequest == 0)               // extender version
        RETLONG(EXTENDERVERSION);

      if (nRequest == 1)               // total functions
        uType = FUNCTION;
      else if (nRequest == 2)          // total constants
        uType = CONSTANT;
      else
        RETERROR(-1, 101);             // Invalid request

      for (i = 0, uCount = 0; CommandExtTable[i].ident != -1; i++)
        if ((UINT) CommandExtTable[i].type == uType)
          uCount++;

      RETLONG(uCount);
    }

/*---------------------------------------------------------------------------
                 xBaseConvert(s:value, i:oldbase, i:newbase)

  Does base conversions
    value:     number to convert (must be delimited if it contains letters)
    from-base: base that 'value' will be converted FROM
    to-base:   base that 'value' will be converted TO
  ---------------------------------------------------------------------------*/
    case 11:
    {
      LPSTR lpValue;
      LONG  nOldBase, nNewBase, nValue, nReturn;

      lpValue  = lpvvArg[0].lpstr;
      nOldBase = lpvvArg[1].x;
      nNewBase = lpvvArg[2].x;

      if (!(*lpValue))
        RETERROR(-1, 112);             // Invalid 'value'

      if ((nOldBase < 2) || (nOldBase > 36) || (nNewBase < 2) || (nNewBase > 36))
        RETERROR(-1, 111);             // Invalid base

      if (nOldBase == 10)              // decimal -> base
      {
        nValue = MyAtol(lpValue);
        MyLtoa(nValue, szWork, (UINT) nNewBase);
        RETSTRING(szWork);
      }

      // base -> decimal, or, base -> base via intermediate decimal step
      nReturn = Base2Dec(lpValue, (UINT) nOldBase);

      if (nReturn == -2)
        RETERROR(-1, 112);             // Invalid 'value'

      if (nNewBase == 10)              // base -> decimal
        RETLONG(nReturn);

      MyLtoa(nReturn, szWork, (UINT) nNewBase);  // base -> base
      RETSTRING(szWork);
    }

/*---------------------------------------------------------------------------
                             xDiskLabelGet(s:drive)

  Returns volume label of specified drive
    drive: drive letter
  ---------------------------------------------------------------------------*/
    case 13:
    {
      LPSTR  lpDrive;
      static char szFileSpec[7];
      DWORD  dwMaxComponentLen, dwFlags;
      LONG   nLen;
      char   cDrive;
      BOOL   bResult;

      lpDrive = lpvvArg[0].lpstr;

      nLen = lstrlen(lpDrive);

      if ((nLen < 1) || (nLen > 2))
        RETERROR(-1, 131);             // Invalid drive letter

      if ((nLen == 2) && (*(lpDrive + 1) != ':'))
        RETERROR(-1, 131);             // Invalid drive letter

      cDrive = *lpDrive;

      if ((cDrive >= 'a') && (cDrive <= 'z'))
        cDrive += ('A' - 'a');         // convert [a-z] to [A-Z]

      if ((cDrive < 'A') || (cDrive > 'Z'))
        RETERROR(-1, 131);             // Invalid drive letter

      wsprintf((LPSTR) szFileSpec, (LPSTR) "%c:\\", cDrive);
      szWork[0] = '\0';
      bResult = GetVolumeInformation(szFileSpec, szWork, sizeof(szWork), NULL, &dwMaxComponentLen, &dwFlags, NULL, 0);

      if ((bResult) && (szWork[0]))    // success
        RETSTRING(szWork);

      RETSTRING("");                   // if we got here, there's no label
    }

/*---------------------------------------------------------------------------
                              xDriveReady(s:drive)

  Checks whether specified drive is ready
    drive: drive letter
  ---------------------------------------------------------------------------*/
    case 14:
    {
      WIN32_FIND_DATA ffd;
      HANDLE          hSearch;
      LPSTR           lpTestFile = "?:\\NUL";  // NUL dev usually "exists" in each dir
      LPSTR           lpDrive;
      DWORD           dwError;
      LONG            nLen, nReturn;
      UINT            uPrevErrorMode;
      char            cDrive;

      lpDrive = lpvvArg[0].lpstr;

      nLen = lstrlen(lpDrive);

      if ((nLen < 1) || (nLen > 2))
        RETERROR(-1, 141);             // Invalid drive letter

      if ((nLen == 2) && (*(lpDrive + 1) != ':'))
        RETERROR(-1, 141);             // Invalid drive letter

      cDrive = MyToUpper(*lpDrive);

      if (!(IsCharAlpha(cDrive)))
        RETERROR(-1, 141);             // Invalid drive letter

      lpTestFile[0] = cDrive;          // fill in the drive letter
      nReturn = 0;                     // assume failure
      uPrevErrorMode = SetErrorMode(SEM_FAILCRITICALERRORS);  // suppress dialog
      hSearch = FindFirstFile(lpTestFile, &ffd);
      dwError = GetLastError();

      if (hSearch != INVALID_HANDLE_VALUE)  // the file exists
      {
        FindClose(hSearch);
        nReturn = 1;                   // so, the drive is ready
      }
      else if (dwError == ERROR_FILE_NOT_FOUND)  // file not found (it happens)
        nReturn = 1;                   // still, this means the drive is ready

      SetErrorMode(uPrevErrorMode);    // restore previous error mode
      RETLONG(nReturn);
    }

/*---------------------------------------------------------------------------
                         xGetElapsed(s:time1, s:time2)

  Calculates difference between two values obtained with GetExactTime()
    time1: ending time
    time2: start time
  ---------------------------------------------------------------------------*/
    case 15:
    {
      struct ExactTime
      {
        char hr[3];
        char mn[3];
        char sc[3];
        char hu[3];
      } etTime[2];

      LPSTR lpTime1, lpTime2;
      UINT  i;
      int   hr, mn, sc, hu;

      lpTime1 = lpvvArg[0].lpstr;
      lpTime2 = lpvvArg[1].lpstr;

      if (lstrlen(lpTime1) != 11)
        RETERROR(-1, 151);             // time1 is invalid

      if (lstrlen(lpTime2) != 11)
        RETERROR(-1, 152);             // time2 is invalid

      lstrcpy((LPSTR) &etTime[0], lpTime1);  // end time
      lstrcpy((LPSTR) &etTime[1], lpTime2);  // start time

      // Input format: "hh:mm:ss:uu".  Replace all ':' with NULL's
      for (i = 0; i < 2; i++)
      {
        etTime[i].hr[2] = '\0';
        etTime[i].mn[2] = '\0';
        etTime[i].sc[2] = '\0';
        etTime[i].hu[2] = '\0';
      }

      hr = (int) (MyAtol(etTime[0].hr) - MyAtol(etTime[1].hr));
      mn = (int) (MyAtol(etTime[0].mn) - MyAtol(etTime[1].mn));
      sc = (int) (MyAtol(etTime[0].sc) - MyAtol(etTime[1].sc));
      hu = (int) (MyAtol(etTime[0].hu) - MyAtol(etTime[1].hu));

      if (hu < 0)                      // borrow from the seconds column
      {
        hu += 100;
        sc--;
      }

      if (sc < 0)                      // borrow from the minutes column
      {
        sc += 60;
        mn--;
      }

      if (mn < 0)                      // borrow from the hours column
      {
        mn += 60;
        hr--;
      }

      if (hr < 0)
        hr += 24;

      // build the output string in the same format as the input string
      wsprintf((LPSTR) szWork, (LPSTR) "%02d:%02d:%02d:%02d", hr, mn, sc, hu);
      RETSTRING(szWork);
    }

/*---------------------------------------------------------------------------
                                 xHex(s:hexval)

  Converts from hex to decimal
    hex-val: hex number to convert (must be delimited if it contains letters)
  Like xBaseConvert, but simpler way to perform the most useful conversion
  ---------------------------------------------------------------------------*/
    case 16:
    {
      LPSTR lpValue;
      LONG  nReturn;

      lpValue = lpvvArg[0].lpstr;

      if (!(*lpValue))
        RETERROR(-1, 161);             // Invalid 'hex-val'

      nReturn = Base2Dec(lpValue, 16);

      if (nReturn == -2)
        RETERROR(-1, 161);             // Invalid 'hex-val'

      RETLONG(nReturn);
    }

/*---------------------------------------------------------------------------
                    xMessageBox(s:title, s:text, i:style)

  The "real" MessageBox function
    title: title of message box
    text:  text to go in message box
    style: type of icon and/or button(s) to use (the API documentation helps)
  Returns the value assigned to the button that was pressed
  ---------------------------------------------------------------------------*/
    case 18:
    {
      LPSTR lpTitle, lpText;
      LONG  nStyle;
      int   nReturn;

      lpTitle = lpvvArg[0].lpstr;
      lpText  = lpvvArg[1].lpstr;
      nStyle  = lpvvArg[2].x;

      if ((bWin98) || (bWin50))        // hack to make the message box come to the foreground
      {
        DWORD_PTR dwPrevTimeout;       // 64-bit port safety.
        BOOL  bSetTimeout;

        dwPrevTimeout = 0;
        bSetTimeout = SystemParametersInfo(SPI_GETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID) &dwPrevTimeout, 0);

        if ((bSetTimeout) && (dwPrevTimeout))
          bSetTimeout = SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID) 0, 0);

        SetForegroundWindow(hWnd);

        if ((bSetTimeout) && (dwPrevTimeout))
          SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID) dwPrevTimeout, 0);
      }

      nReturn = MessageBox(hWnd, lpText, lpTitle, (UINT) nStyle);
      RETLONG(nReturn);
    }

/*---------------------------------------------------------------------------
                            xVerifyCCard(s:cardnum)

  Verifies a credit card number
    cardnum: credit card number, can include spaces or hyphens
  Returns TRUE if card number is valid, FALSE if it isn't
  ---------------------------------------------------------------------------*/
    case 19:
    {
      LPSTR  lpCardNum, lpSrc;
      static char szCard[MAXCARDLEN + 1];
      UINT   i, uLen, uDigit, uCheckVal;
      BOOL   bOddEven;

      lpCardNum = lpvvArg[0].lpstr;

      if (!(*lpCardNum))
        RETLONG(FALSE);

      lpSrc = lpCardNum;

      // extract letters and numbers; skip over spaces, hyphens, etc.
      for (i = 0; (i < MAXCARDLEN) && (*lpSrc); lpSrc++)
        if (IsCharAlphaNumeric(*lpSrc))
          szCard[i++] = *lpSrc;

      szCard[i] = '\0';
      uLen = i;
      bOddEven = uLen % 2;

      // Credit cards are typically 13, 15, or 16 characters long, but this
      // routine doesn't check for length.  Hence, any number that uses this
      // algorithm (such as checking account numbers) will pass our test.

      for (i = uCheckVal = 0; i < (uLen - 1); i++)
      {
        uDigit = szCard[i] - '0';

        if (i % 2 == (UINT) bOddEven)
          uDigit *= 2;

        if (uDigit >= 10)
          uDigit -= 9;

        uCheckVal += uDigit;
      }

      i = 10 - (uCheckVal % 10);

      if (i == 10)
        i = 0;

      if (i == (UINT) (szCard[uLen - 1] - '0'))
        RETLONG(TRUE);

      RETLONG(FALSE);
    }

/*---------------------------------------------------------------------------
            xGetChildHwnd(i:parent-hwnd, s:child-text, i:child-seq)

  Gets the handle of a child window belonging to the specified parent.
    parent-hwnd: handle of the parent window
    child-text:  text string of the child window (may be blank)
    child-seq:   positional sequence of the child window (may be 0)
  Either child-text or child-seq, or both, must be specified.  If only the
  text is specified, it will be compared to all children until a match is
  found; if not, 0 is returned.  If only the sequence is specified, it will
  be assumed to be correct; if there are less than 'seq' children, 0 is
  returned.  If both text and sequence are specified, the text is compared
  to only that one specific child; if they don't match, 0 is returned.
  Text comparisons are case-insensitive, the '&' in the control string (if
  any) is stripped out, but leading and trailing spaces are not stripped.
  ---------------------------------------------------------------------------*/
    case 20:
    {
      HWND   hParentWnd, hWndChild;
      LPSTR  lpChildText, lpText, lpWork;
      static char szText[256], szSeqBuf[16];
      LONG   nChildSeq;
      UINT   i, uSeq;
      BOOL   bFound;

      hParentWnd  = LongToHandle(lpvvArg[0].x);  // 64-bit port - allow sign extension.
      lpChildText = lpvvArg[1].lpstr;
      nChildSeq   = lpvvArg[2].x;

      if ((!(*lpChildText)) && (nChildSeq == 0))  // text is "" and seq == 0
        RETERROR(-1, 203);             // Must specify child text and/or seq #

      if (!(IsWindow((HWND) hParentWnd)))
        RETERROR(-1, 204);             // Parent window doesn't exist

      if (nChildSeq == 0)              // no sequence number was specified
        uSeq = 1;                      // so we set it to be the current seq
      else
        uSeq = (UINT) nChildSeq;       // the user told us when to stop

      // get first child belonging to the specified parent
      hWndChild = GetWindow((HWND) hParentWnd, GW_CHILD);

      if ((uDebugLevel >= 1) && (!(hWndChild)))
        DebugMessage(hWnd, "No child windows", "GW_CHILD failed", bDebugMonitor);

      bFound = FALSE;

      // If the preceding GetWindow call (get first child) failed, we will
      // not enter this loop.  Otherwise, we will step through the chain of
      // siblings until we run out of windows or until we find the one matching
      // the specified text.  Or, if the user specified a specific child
      // sequence number, we will go directly to that and ignore the others.

      for (i = 1; (i <= uSeq) && (hWndChild); i++)
      {
        if ((*lpChildText) && (i == uSeq))  // we're interested in this text
        {
          GetWindowText(hWndChild, szWork, sizeof(szWork));

          lpWork = szWork;
          lpText = szText;

          while (*lpWork)
          {
            if (*lpWork == '&')        // strip any '&' from control text
              lpWork++;
            else
            {
              if (IsDBCSLeadByte(*lpWork))  // if it's the lead byte of a double-byte character
                *lpText++ = *lpWork++;  // we need to copy an extra byte

              *lpText++ = *lpWork++;
            }
          }

          *lpText = '\0';

          if (uDebugLevel >= 2)
          {
            // display the current sequence number and the control text
            wsprintf((LPSTR) szSeqBuf, (LPSTR) "Sequence: %u", i);

            if (!(DebugMessage(hWnd, szText, szSeqBuf, bDebugMonitor)))
              RETCANCEL;               // the user pressed "Cancel"
          }

          if (lstrcmpi(szText, lpChildText) == 0)  // we have a match
          {
            bFound = TRUE;
            break;                     // express exit out of the 'for' loop
          }

          if (nChildSeq != 0)          // a match was required for this seq
            break;                     // and we didn't get one
        }

        if (nChildSeq == 0)            // no sequence number was specified
          uSeq = i + 1;                // 'i' will == 'seq' at next iteration
        else if (i == uSeq)
        {
          bFound = TRUE;               // we just processed the desired seq
          break;                       // so our work here is done
        }

        // get next sibling of the current child
        hWndChild = GetWindow(hWndChild, GW_HWNDNEXT);

        if ((uDebugLevel >= 1) && (!(hWndChild)))
        {
          static char szWork2[256];

          LoadMyExtString(hinstDLL, 7001, szWork, sizeof(szWork));  // "GW_HWNDNEXT failed"
          LoadMyExtString(hinstDLL, 7002, szWork2, sizeof(szWork2));  // "Last child window"
          DebugMessage(hWnd, szWork2, szWork, bDebugMonitor);
        }
      }

      if (bFound)
        RETLONG((LONG) HandleToLong(hWndChild));     // this was the hWnd at the 'break' (64-bit port silence.)

      RETLONG(0);                      // return 0 on failure
    }

/*---------------------------------------------------------------------------
                xSendMessage(i:hwnd, i:msg, i:wparam, i:lparam)

  The real SendMessage.
    hwnd:   window handle
    msg:    message number
    wparam: first message parameter
    lparam: second message parameter (must be numeric)
  Returns the value returned by SendMessage
  ---------------------------------------------------------------------------*/
    case 21:
    {
      LONG nMsg, nWparam, nLparam, nReturn;
      HWND hWnd;

      hWnd    = LongToHandle(lpvvArg[0].x); // 64-bit port silence.
      nMsg    = lpvvArg[1].x;
      nWparam = lpvvArg[2].x;
      nLparam = lpvvArg[3].x;

      if (!(IsWindow(hWnd)))
        RETERROR(-1, 215);             // Specified window doesn't exist

      nReturn = (LONG) SendMessage(hWnd, (UINT) nMsg, (WPARAM) nWparam, (LPARAM) nLparam);
      RETLONG(nReturn);
    }

/*---------------------------------------------------------------------------
                             xEjectMedia(s:drive)

  Ejects removable media from a device.
    drive: drive letter of device (eg, "G", or "G:")
  Returns TRUE on success, FALSE on failure
  ---------------------------------------------------------------------------*/
    case 22:
    {
      HANDLE hDevice;
      LPSTR  lpDrive;
      DWORD  dwBytesReturned;
      UINT   uPrevErrorMode;
      char   cDrive;
      BOOL   bReturn;

      lpDrive = lpvvArg[0].lpstr;

      if (!(*lpDrive))
        RETERROR(-1, 221);             // Invalid drive letter

      cDrive = MyToUpper(*lpDrive);

      if (!(IsCharAlpha(cDrive)))
        RETERROR(-1, 221);             // Invalid drive letter

      if (bWinNT)                      // Windows NT
      {
        static char szDrivePath[7];
        DWORD  dwError;

        wsprintf((LPSTR) szDrivePath, (LPSTR) "\\\\.\\%c:", cDrive);

        uPrevErrorMode = SetErrorMode(SEM_FAILCRITICALERRORS);  // suppress dialog
        hDevice = CreateFile(szDrivePath, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
        SetErrorMode(uPrevErrorMode);  // restore previous error mode

        if (hDevice == INVALID_HANDLE_VALUE)
          RETLONG(0);

        bReturn = DeviceIoControl(hDevice, IOCTL_STORAGE_EJECT_MEDIA, NULL, 0, NULL, 0, &dwBytesReturned, NULL);

        if (!(bReturn))
        {
          dwError = GetLastError();

          if (dwError == ERROR_MEDIA_CHANGED)  // unimportant error
            bReturn = TRUE;
        }

        CloseHandle(hDevice);
        RETLONG(bReturn);
      }
      else if (bWin95)                 // Windows 95
      {
        DIOC_REGISTERS reg;
        UINT           uAttempts;
        int            nDrive;

        hDevice = CreateFile("\\\\.\\VWIN32", 0, 0, NULL, 0, FILE_FLAG_DELETE_ON_CLOSE, NULL);

        if (hDevice == INVALID_HANDLE_VALUE)
          RETLONG(0);

        nDrive = cDrive - 'A' + 1;     // A=1, B=2, etc.

        for (uAttempts = 0; uAttempts < 5; uAttempts++)  // retry up to 5 times
        {
          reg.reg_EAX = 0x440D;        // IOCTL for block devices
          reg.reg_EBX = nDrive;        // zero-based drive identifier
          reg.reg_ECX = 0x0849;        // Eject Media command
          reg.reg_Flags = 0x0001;      // assume error (carry flag is set)

          bReturn = DeviceIoControl(hDevice, VWIN32_DIOC_DOS_IOCTL, &reg, sizeof(reg), &reg, sizeof(reg), &dwBytesReturned, NULL);

          if (!(bReturn))              // function failed
            break;

          if (reg.reg_Flags & 0x0001)  // error if carry flag is set
          {
            bReturn = FALSE;

            if (reg.reg_EAX == 0x05)   // "Access denied" -- may indicate the device isn't ready
              Sleep(1000);             // wait one second (and then retry)
            else                       // some other error
              break;                   // don't retry
          }
          else                         // success
            break;
        }

        CloseHandle(hDevice);
        RETLONG(bReturn);
      }
      else                             // some other 32-bit version of Windows
        RETERROR(-1, 224);             // Not supported in this Windows version
    }


/*---------------------------------------------------------------------------
                                  default case
  ---------------------------------------------------------------------------*/
    default:                           // we shouldn't get here
    {
      RETERROR(-1, 901);               // Unknown function
    }
  }                                    // end of "big switch"
}

/*===========================================================================
                                     MyAtol

  Converts a decimal character string to a long integer.
  ===========================================================================*/

LONG NEARPASCAL MyAtol(LPSTR lpSource)
{
  LPSTR lpSrc;
  LONG  nNum;
  BOOL  bNegate;

  lpSrc = lpSource;
  bNegate = FALSE;
  nNum = 0;

  while (*lpSrc == ' ')
    lpSrc++;

  if (*lpSrc == '-' || *lpSrc == '+')
  {
    if (*lpSrc == '-')
      bNegate = TRUE;

    lpSrc++;
  }

  while (*lpSrc >= '0' && *lpSrc <= '9')
  {
    nNum = nNum * 10 + (*lpSrc - '0');
    lpSrc++;
  }

  if (bNegate)
    nNum = -nNum;

  return (nNum);
}

/*===========================================================================
                                     MyLtoa

  Converts a decimal integer to a string in any base (2 >= base >= 36).
  ===========================================================================*/

void NEARPASCAL MyLtoa(LONG nNum, LPSTR lpDest, UINT nBase)
{
  static char szConvert[37] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  static char szNum[30];
  LONG   nTemp;
  UINT   i, j;

  if (nNum == 0x80000000)
  {
    lstrcpy(lpDest, "-2147483648");
    return;
  }

  i = 0;
  j = 29;

  if (nNum < 0)
  {
    lpDest[i++] = '-';
    nNum = -nNum;
  }

  szNum[j--] = 0;

  do
  {
    nTemp = nNum % nBase;
    szNum[j--] = szConvert[(UINT) nTemp];
    nNum = nNum / nBase;
  }
  while(nNum);

  j++;
  lstrcpy(lpDest + i, szNum + j);

  return;
}

/*===========================================================================
                                  DebugMessage

  Displays a debug message in a MessageBox or on a debug monitor.
  ===========================================================================*/

BOOL NEARPASCAL DebugMessage(HWND hWnd, LPSTR lpText, LPSTR lpTitle, BOOL bDebugMonitor)
{
  int  nReturn;

  if (bDebugMonitor)
  {
    OutputDebugString("\r\n");
    OutputDebugString(lpTitle);
    OutputDebugString("\r\n");
    OutputDebugString(lpText);
    OutputDebugString("\r\n");
    return (TRUE);
  }

  nReturn = MessageBox(hWnd, lpText, lpTitle, MB_ICONINFORMATION | MB_OKCANCEL);

  if (nReturn == IDCANCEL)
    return (FALSE);

  return (TRUE);
}

/*===========================================================================
                                    Base2Dec

  Converts a string in any base (2 >= base >= 36) to a decimal integer.
  ===========================================================================*/

LONG NEARPASCAL Base2Dec(LPSTR lpBaseVal, UINT uBase)
{
  LPSTR lpPos;
  char  cChar;
  LONG  nColumn, nTot;
  UINT  i, uLen, uMaxChar;

  if ((uBase < 2) || (uBase > 36))
    return (-1);

  uMaxChar = uBase - 1;                // highest legal character in this base

  uLen = lstrlen(lpBaseVal);
  lpPos = lpBaseVal + uLen - 1;        // pointer to last character in string

  nTot = 0;

  for (i = 0; i < uLen; i++, lpPos--)
  {
    cChar = *lpPos;

    if ((cChar >= 'a') && (cChar <= 'z'))
      cChar += ('A' - 'a');            // convert [a-z] to [A-Z]

    if ((cChar >= 'A') && (cChar <= 'Z'))
      cChar -= ('0' + ('A' - '9' - 1));  // convert [A-Z] to [10-36]
    else
      cChar -= '0';                    // convert numeric character to integer

    if ((cChar < 0) || ((UINT) cChar > uMaxChar))  // check for bad character
      return (-2);                     // illegal character

    nColumn = Pow(uBase, i);           // calculate the value of this "column"
    nTot += (cChar * nColumn);
  }

  return (nTot);
}

/*===========================================================================
                                      Pow

  Raises an integer to the specified power (x ** y).
  ===========================================================================*/

LONG NEARPASCAL Pow(UINT uBase, UINT uExponent)
{
  LONG nTot;
  UINT i;

  if (uExponent == 0)                  // (x ** 0) is always 1
    return (1);

  nTot = uBase;

  for (i = 1; i < uExponent; i++)
    nTot *= uBase;

  return (nTot);
}

/*===========================================================================
                                 LoadMyExtString

  This allows for localization of error messages and other strings.
  ===========================================================================*/

int NEARPASCAL LoadMyExtString(HMODULE hInst, int wID, LPSTR lpBuffer, int nBuffermax)
{
  static char szLangSection[] = "WILX Extender";
  LPSTR       lpSlash;
  int         nCharsBack, i;
  static char szLangIni[MAX_PATH];
  static int  nIniExist = 0;           // 0=uninit   1=notrans   2=xlate it
  static char szLANGDLL[] = "WWWELANG";
  static char szXXX[] = "XXX";
  static char szMAIN[] = "MAIN";
  static char szINTL[] = "INTL";
  static char szLibIni[] = "WWWBATCH.INI";
  char        SafeBuffer[20];
  char        szKey[20];

  if (nIniExist == 0 || wID == 0)      // Process not initialized
  {
    OFSTRUCT OFS;
    HFILE    hRes;

    nIniExist = 1;
    szLangIni[0] = '\0';

    // First see if a language override has been installed
    GetPrivateProfileString(szMAIN, "Language", szXXX, SafeBuffer, sizeof(SafeBuffer), szLibIni);

    if (lstrcmp(SafeBuffer, szXXX) == 0)  // no language override
      GetProfileString(szINTL, "sLanguage", "enu", SafeBuffer, sizeof(SafeBuffer));  // look foreign lang.  Def to enu

    // Get DLL location for search for our lng dll
    i = GetModuleFileName(hInst, szLangIni, sizeof(szLangIni));

    if (i == 0)
      szLangIni[0] = '\0';
    else
    {
      lpSlash = MyLastSlashInString(szLangIni);

      if (lpSlash)
        *(AnsiNext(lpSlash)) = '\0';
      else
        szLangIni[0] = '\0';
    }

    lstrcat(szLangIni, szLANGDLL);     // glue on WWWELANG
    lstrcat(szLangIni, ".");           // glue on .
    lstrcat(szLangIni, SafeBuffer);    // glue on enu or whatnot

    hRes = (HFILE) OpenFile(szLangIni, &OFS, OF_EXIST);  // well, does it exist?

    if (hRes == (HFILE) -1)
    {
      wsprintf((LPSTR) szLangIni, (LPSTR) "%s.%s", (LPSTR) szLANGDLL, (LPSTR) SafeBuffer);
      hRes = (HFILE) OpenFile(szLangIni, &OFS, OF_EXIST);

      if (hRes != (HFILE) -1)
        lstrcpyn(szLangIni, OFS.szPathName, sizeof(szLangIni));
    }

    if (hRes != (HFILE) -1)
      nIniExist = 2;
  }

  if (wID >= 100 && nIniExist == 2)
  {
    szKey[0] = 'X';
    szKey[1] = '\0';
    MyLtoa(wID, szKey + 1, 10);

    nCharsBack = GetPrivateProfileString(szLangSection, szKey, szXXX, lpBuffer, nBuffermax, szLangIni);

    if (lstrcmp(lpBuffer, szXXX) == 0)
    {
      lpBuffer[0] = '\0';
      nCharsBack = LoadString(hInst, wID, lpBuffer, nBuffermax);
    }
  }
  else
  {
    lpBuffer[0] = '\0';
    nCharsBack = LoadString(hInst, wID, lpBuffer, nBuffermax);
  }

  while (*lpBuffer)
  {
    if (*lpBuffer == '~')
      *lpBuffer = '\n';

    if (*lpBuffer == '|')
      *lpBuffer = '\t';

    lpBuffer = AnsiNext(lpBuffer);
  }

  return nCharsBack;
}

/*===========================================================================
                                MyLastSlashInString

  Returns a pointer to the last slash (forward or backward) found in a string.
  Supports DBCS.
  ===========================================================================*/

LPSTR NEARPASCAL MyLastSlashInString(LPSTR lpString)
{
  LPSTR lpPtr, lpSlash;

  lpSlash = NULL;

  for (lpPtr = lpString; *lpPtr; lpPtr = AnsiNext(lpPtr))
    if ((*lpPtr == '\\') || (*lpPtr == '/'))
      lpSlash = lpPtr;

  return (lpSlash);
}
